/*CAMERA.C********************************************************************

   Clax: Portable keyframing library
         see inclosed LICENSE.TXT for licensing terms.

         for documentation, refer to CLAX.TXT

   author           : Adept
   file created     : 16/04/97
   file description : camera math

   revision history :
     v0.10 (16/04/97) Adept: Initial version.

   notes            :

*****************************************************************************/

#include <math.h>
#include <stdio.h>
#include "clax.h"
#include "claxi.h"

void cam_lens_fov (float lens, float *fov)
{
/*
  cam_lens_fov: converts 3dstudio lens to fov.
*/
  int i;
  struct {
    float lens, fov;
  } lens_table[] = {
    {15.0,  115.0}, {20.0, 94.28571}, {24.0, 84.0}, {28.0,  76.36364},
    {35.0,  63.0},  {50.0, 46.0},     {85.0, 28.0}, {135.0, 18.0},
    {200.0, 12.0}
  };

  for (i = 0; i < 9; i++) if (lens == lens_table[i].lens) {
    *fov = lens_table[i].fov;
    return;
  }
  *fov = 15.0 / lens * 160;
}

void cam_update (c_CAMERA *cam)
{
/*
  cam_matrix: creates a camera matrix. 
*/
  c_VECTOR c;
  float    tanfov, focus;
  float    ax, ay, az;
  float    sinx, siny, sinz,
           cosx, cosy, cosz;

  vec_sub (&cam->target, &cam->pos, &c);
  focus = vec_length (&c);
  tanfov = tan ((cam->fov / 2) / 180 * M_PI);
  cam->perspX = cam->sizeX / 2 / tanfov;
  cam->perspY = cam->sizeY / 2 / cam->aspectratio / tanfov;

  ax = -atan2 (c.x, c.z);
  ay = asin (c.y / focus);
  az = -cam->roll * M_PI / 180;

  sinx = sin (ax); cosx = cos (ax); siny = sin (ay);
  cosy = cos (ay); sinz = sin (az); cosz = cos (az);

  mat_identity (cam->camera_mat);
  cam->camera_mat[X][X] =  sinx * siny * sinz + cosx * cosz;
  cam->camera_mat[X][Y] =  cosy * sinz;
  cam->camera_mat[X][Z] =  sinx * cosz - cosx * siny * sinz;
  cam->camera_mat[Y][X] =  sinx * siny * cosz - cosx * sinz;
  cam->camera_mat[Y][Y] =  cosy * cosz;
  cam->camera_mat[Y][Z] = -cosx * siny * cosz - sinx * sinz;
  cam->camera_mat[Z][X] = -sinx * cosy;
  cam->camera_mat[Z][Y] =  siny;
  cam->camera_mat[Z][Z] =  cosx * cosy;
}
