{
                             
                                           
                                        
                                                    
                                       
                             
                                    
                                    
                               
}
{$A+,B-,D-,E-,F+,G+,I-,L-,N+,O-,P-,Q-,R-,S-,T-,V+,X+}
Unit TXMS;
interface
var
  XMSErrorCode: Byte;  { Error Code - Defined in XMS Spec }
  XMSAddr: pointer;    { Entry Point for HIMEM.SYS Driver }

function XMSDriverLoaded: Boolean;
function XMSTotalMemoryAvailable: Word;
function XMSLargestBlockAvailable: Word;
function XMSAllocateBlock(KBSize: Word): Word;
function XMSReleaseBlock(Handle: Word): Boolean;
function XMSMoveDataTo(SourceAddr: pointer; NumBytes: Longint; XMSHandle: Word; XMSOffset: Longint): Boolean;
function XMSGetDataFrom(XMSHandle: Word; XMSOffset: Longint; NumBytes: Longint; LowMemAddr: pointer): Boolean;
Procedure XMSFillBlock(Handle,KByte:Word;F:Byte);
implementation
type
  XMSMoveStruct = record
                    movelen: Longint;  { length of block to move in bytes }
                    case Integer of
                      { Case 0 Variant for Low Memory to XMS }
                      0: (SHandle: Word;        { Source Handle = 0 for conventional memory }
                          SPtr: pointer;        { Source address }
                          XMSHdl: Word;         { XMS destination handle }
                          XMSOffset: Longint);  { 32 bit XMS offset }
                      { Case 1 Variant for XMS to Low Memory }
                      1: (XMSH: Word;           { XMS source handle }
                          XMSOfs: Longint;      { starting offset in XMS }
                          DHandle: Word;        { 0 when conventional memory destination }
                          DPtr: pointer);       { address in conventional memory }
                    end;

var
  moveparams: XMSMoveStruct;  { structure for moving to and from XMS }
  HandleNum : Word;
  HandleAr  : Array [1..1000] Of Word;
  RealExitP : Pointer;
{***************************************************************************}
{ XMSDriverLoaded - Returns true if Extended Memory Driver HIMEM.SYS Loaded }
{                 - Sets Entry Point Address: XMSAddr                       }
{***************************************************************************}
function XMSDriverLoaded: Boolean;
const
  himemseg: Word = 0;
  himemofs: Word = 0;
begin
  XMSErrorCode := 0;
  asm
    mov ax,4300h  { Check to see if HIMEM.SYS installed }
    int 2fh
    cmp al,80h    { Returns AL = 80h if installed }
    jne @1
    mov ax,4310h  { Now get the entry point }
    int 2fh
    mov himemofs,bx
    mov himemseg,es
  @1:
  end;
  XMSDriverLoaded := himemseg <> 0;
  XMSAddr := ptr(himemseg, himemofs);
end;

{**************************************************************}
{ XMSTotalMemoryAvailable - Returns Total XMS Memory Available }
{**************************************************************}
function XMSTotalMemoryAvailable: Word;
begin
  XMSErrorCode := 0;
  XMSTotalMemoryAvailable := 0;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;

  asm
    mov  ah,8
    call XMSAddr
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { Set Error Code }
    xor  dx,dx
  @1:
    mov  @Result,dx       { DX = total free extended memory }
  end;
end;

{****************************************************************************}
{ XMSLargestBlockAvailable - Returns Largest Continguous XMS Block Available }
{****************************************************************************}
function XMSLargestBlockAvailable: Word;
begin
  XMSErrorCode := 0;
  XMSLargestBlockAvailable := 0;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;

  asm
    mov  ah,8
    call XMSAddr
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { On Error, Set Error Code }
  @1:
    mov  @Result,ax       { AX = largest free XMS Block }
  end;
end;

{**************************************************************}
{ XMSAllocateBlock - Allocates Block of XMS Memory             }
{                  - Input: KBSize: No. of Kilobytes requested }
{                  - Return Handle for memory if successful    }
{**************************************************************}
function XMSAllocateBlock(KBSize: Word): Word;
Var Res:Word;
begin
  XMSAllocateBlock := 0;
  XMSErrorCode := 0;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;

  asm
    mov  ah,9
    mov  dx,KBSize
    call XMSAddr
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { On Error, Set Error Code }
    xor  dx,dx
  @1:
    mov Res,dx        { DX = handle for extended memory }
  end;
  Inc(HandleNum);
  HandleAr[HandleNum]:=Res;
  XMSAllocateBlock:=Res;
end;

{*******************************************************************}
{ XMSReleaseBlock - Releases Block of XMS Memory                    }
{                 - Input: Handle identifying memory to be released }
{                 - Returns true if succesful                       }
{*******************************************************************}
function XMSReleaseBlock(Handle: Word): Boolean;
var
  OK,LP: Word;
begin
  XMSErrorCode := 0;
  XMSReleaseBlock := false;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;

  If HandleNum=0 then Exit;{Should never happend!!!}
  Ok:=1;
  While (HandleAr[Ok]<>Handle) And (Ok<=HandleNum) Do Inc(Ok);
  If Ok<HandleNum then
   For Lp:=Ok+1 To HandleNum Do HandleAr[Lp-1]:=HandleAr[Lp];
  Dec(HandleNum);

  asm
    mov  ah,0Ah
    mov  dx,Handle
    call XMSAddr
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { On Error, Set Error Code }
  @1:
    mov OK,ax
  end;
  XMSReleaseBlock := OK <> 0;
end;

{****************************************************************************}
{ XMSMoveDataTo - Moves Block of Data from Conventional Memory to XMS Memory }
{               - Data Must have been previously allocated                   }
{               - Input - SourceAddr: address of data in conventional memory }
{                       - NumBytes:   number of bytes to move                }
{                       - XMSHandle:  handle of XMS block                    }
{                       - XMSOffset:  32 bit destination offset in XMS Block }
{               - Returns true if completed successfully                     }
{****************************************************************************}
function XMSMoveDataTo(SourceAddr: pointer; NumBytes: Longint; XMSHandle: Word; XMSOffset: Longint): Boolean;
var
  Status: Word;
begin
  XMSErrorCode := 0;
  XMSMoveDataTo := false;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;
  MoveParams.MoveLen   := NumBytes;
  MoveParams.SHandle   := 0; { Source Handle = 0 For Conventional Memory }
  MoveParams.SPtr      := SourceAddr;
  MoveParams.XMSHdl    := XMSHandle;
  MoveParams.XMSOffset := XMSOffset;

  asm
    mov  ah,0Bh
    mov  si,offset MoveParams
    call XMSAddr
    mov  Status,ax        { Completion Status }
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { Save Error Code }
  @1:
  end;
  XMSMoveDataTo := Status <> 0;
end;

{******************************************************************************}
{ XMSGetDataFrom - Moves Block From XMS to Conventional Memory                 }
{                - Data Moust have been previously allocated and moved to XMS  }
{                - Input - XMSHandle:  handle of source XMS Block              }
{                        - XMSOffset:  32 bit source offset in XMS Block       }
{                        - NumBytes:   number of bytes to move                 }
{                        - LowMemAddr: destination addr in conventional memory }
{               - Returns true if completed successfully                       }
{******************************************************************************}
function XMSGetDataFrom(XMSHandle: Word; XMSOffset: Longint; NumBytes: Longint; LowMemAddr: pointer): Boolean;
var
  Status: Word;
begin
  XMSErrorCode := 0;
  XMSGetDataFrom := false;
  if XMSAddr = nil then  { Check if HIMEM.SYS Loaded }
    if not XMSDriverLoaded then exit;
  MoveParams.MoveLen := NumBytes;  { Set-Up Structure to Pass }
  MoveParams.XMSH    := XMSHandle;
  MoveParams.XMSOfs  := XMSOffset;
  MoveParams.DHandle := 0;         { Dest Handle = 0 for Conventional Memory }
  MoveParams.DPtr    := LowMemAddr;

  asm
    mov  ah,0Bh
    mov  si,offset MoveParams
    call XMSAddr
    mov  Status,ax        { Completion Status }
    or   ax,ax
    jnz  @1
    mov  XMSErrorCode,bl  { Save Error Code }
  @1:
  end;
  XMSGetDataFrom := Status <> 0;
end;


{****************************************************************************}
{$F+}{                                                                       }
{Procedure TXMSExitProc - Auto Relase all Handles if user hadn't released    }
{                         them . This is internal procedure . User should    }
{                         have no effect on her .                            }
{                                                                            }
{****************************************************************************}
Procedure TXMSExitProc;
Var Lp:Word;
Begin
 ExitProc:= RealExitP;
 If HandleNum=0 then Exit;
 For Lp:=1 to HandleNum Do XMSReleaseBlock(HandleAr[Lp]);
End;
{****************************************************************************}
{ Procedure XMSFillBlock - Fill XMS memory with specified value .            }
{                                                                            }
{****************************************************************************}
Procedure XMSFillBlock(Handle,KByte:Word;F:Byte);
Var Ar:Array[1..1024] of byte;
    A:Word;
Begin
 FillChar(Ar,1024,F);
 For A:=0 to KByte-1 Do XMSMoveDataTo(@Ar,1024,Handle,A shl 10);
End;

begin
  RealExitP := ExitProc;
  ExitProc := @TXMSExitProc;
  XMSAddr := nil;  { Initialize XMSAddr }
  XMSErrorCode := 0;
  HandleNum := 0;
end.