// mighty.engine version 0.10
// internal routines

#include <math.h>
#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

int       MAXX;
int       MAXY;
int       CLIPX;
int       CLIPY;
int       XPROJ;
int       YPROJ;
int       asine_table [256];
char      *VirtualScreen;
char      *ActiveTexture;
char      *LightTable;
Polygon   **sortbuffer;
Point     *tPoint;              // temporary point
Scene     *scene;               // active scene

void InitEngine (int xRes, int yRes)
{
  int i;

  for (i = -128; i < 128; i++)
    asine_table [i+128] = int (16383.0*(PI/2+asin (i/128.0))/(PI/1));
  XPROJ = xRes/2;
  YPROJ = yRes/2;
  CLIPX = xRes-1;
  CLIPY = yRes-1;
  MAXX = xRes;
  MAXY = yRes;
  tPoint = new Point (0, 0, 1);
}

void DoneEngine ()
{
  delete tPoint;
}

void Intersect (ClipPoint *src, ClipPoint *dst, ClipPoint *intersect)
{
  float Ratio = (float)(CLIPDISTANCE-dst->z) / (float)(src->z-dst->z);
  intersect->x = (src->x-dst->x)*Ratio + dst->x;
  intersect->y = (src->y-dst->y)*Ratio + dst->y;
  intersect->z = CLIPDISTANCE;
  intersect->c = (src->c-dst->c)*Ratio + dst->c;
  intersect->u1 = (src->u1-dst->u1)*Ratio + dst->u1;
  intersect->v1 = (src->v1-dst->v1)*Ratio + dst->v1;
}

int ZClip (ClipPoint *src, ClipPoint *dst, int code)
{
  int NumPoints;

  switch (code) {
    case 1:
      Intersect (&src[1], &src[0], &dst[0]);
      dst[1] = src[1];
      dst[2] = src[2];
      Intersect (&src[2], &src[0], &dst[3]);
      NumPoints = 4;
      break;
    case 2:
      dst[0] = src[0];
      Intersect (&src[0], &src[1], &dst[1]);
      Intersect (&src[2], &src[1], &dst[2]);
      dst[3] = src[2];
      NumPoints = 4;
      break;
    case 3:
      Intersect (&src[2], &src[1], &dst[0]);
      dst[1] = src[2];
      Intersect (&src[2], &src[0], &dst[2]);
      NumPoints = 3;
      break;
    case 4:
      dst[0] = src[0];
      dst[1] = src[1];
      Intersect (&src[1], &src[2], &dst[2]);
      Intersect (&src[0], &src[2], &dst[3]);
      NumPoints = 4;
      break;
    case 5:
      Intersect (&src[1], &src[0], &dst[0]);
      dst[1] = src[1];
      Intersect (&src[1], &src[2], &dst[2]);
      NumPoints = 3;
      break;
    case 6:
      dst[0] = src[0];
      Intersect (&src[0], &src[1], &dst[1]);
      Intersect (&src[0], &src[2], &dst[2]);
      NumPoints = 3;
  }
  return NumPoints;
}

void SubVec (Point *p1, Point *p2)
{
  tPoint->x3d = p1->x3d - p2->x3d;
  tPoint->y3d = p1->y3d - p2->y3d;
  tPoint->z3d = p1->z3d - p2->z3d;
}

int GouraudLight (Point *a, Point *b)
{
  int   i;
  float tempIntensity, intensity=0.0;

  for (i=0; i < scene->numLights; i++) {
    SubVec (scene->light[i], a);
    tPoint->Normalise ();
    tempIntensity = DotProduct (b, tPoint);
    if (tempIntensity > 0.0) intensity += tempIntensity;
  }
  intensity += scene->AmbientLight;
  if (intensity > 1.0) intensity = 1.0;
  return (char (intensity*scene->MaxIntensity));
}

void qSort (int l, int r)
{
  int i, j, x;
  Polygon *temp;

  i = l; j = r;
  x = sortbuffer [(l+r) >> 1]->midZ;
  do {
    while (sortbuffer [i]->midZ > x) i++;
    while (sortbuffer [j]->midZ < x) j--;
    if (i <= j) {
      temp = sortbuffer [i];
      sortbuffer [i] = sortbuffer [j];
      sortbuffer [j] = temp;
      i++; j--;
    }
  } while (i <= j);
  if (l<j) qSort (l, j);
  if (i<r) qSort (i, r);
}

void quickSort (int size)
{
  qSort (0, size-1);
}