// mighty.engine version 0.10
// keyframer class

#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

Keyframer::Keyframer (int frames)
{
  int i;

  numFrames = frames-1;
  Frame = 0;
  camera = new Fcamera;
  object = new Fobject* [scene->numObjects];
  light = new Flight* [scene->numLights];
  camera->source = new Vector [frames];
  camera->dest = new Vector [frames];
  camera->roll = new float [frames];
  for (i=0; i < scene->numObjects; i++) {
    object [i] = new Fobject;
    object [i]->rotate = new quat [frames];
    object [i]->scale = new Vector [frames];
    object [i]->position = new Vector [frames];
  }
  for (i=0; i < scene->numLights; i++) {
    light [i] = new Flight;
    light [i]->position = new Vector [frames];
  }
}

Keyframer::~Keyframer ()
{
  int i;

  for (i=0; i < scene->numObjects; i++) {
    delete object [i]->rotate;
    delete object [i]->scale;
    delete object [i]->position;
    delete object [i];
  }
  for (i=0; i < scene->numLights; i++) {
    delete light [i]->position;
    delete light [i];
  }
  delete camera->source;
  delete camera->dest;
  delete camera->roll;
  delete camera;
  delete object;
  delete light;
}

void Keyframer::Prepare ()
{
  int i;
  Vector v, bla = {1.0, 1.0, 1.0};

  for (i=0; i < scene->numObjects; i++) {
    scene->object [i]->SetScale (bla);
    scene->object [i]->TransformLocal ();
    scene->object [i]->Save ();
    scene->object [i]->SetPivot (object [i]->pivot);
  }
}

void Keyframer::NextFrame ()
{
  SetFrame (Frame);
  Frame++;
}

void Keyframer::SetFrame (int Frame)
{
  int   i;
  float matrix[9];
  quat  temp;
  scene->camera->SetTo
    (camera->source[Frame].x,
     camera->source[Frame].y,
     camera->source[Frame].z,
     camera->dest[Frame].x,
     camera->dest[Frame].y,
     camera->dest[Frame].z);
  for (i=0; i < scene->numObjects; i++) {
    qt_invmatrix (object [i]->rotate [Frame], matrix);
    scene->object [i]->SetMatrix (matrix);
    scene->object [i]->SetScale (object [i]->scale [Frame]);
    scene->object [i]->SetPosition (object [i]->position [Frame]);
    scene->object [i]->TransformLocal ();
  }
  for (i=0; i < scene->numLights; i++) {
    scene->light [i]->SetCoords
      (light [i]->position [Frame].x,
       light [i]->position [Frame].y,
       light [i]->position [Frame].z);
  }
  Frame++;
}
