// mighty.engine version 0.10
// object class

#include <stdlib.h>
#include <string.h>
#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

Object::Object ()
{
  trans.x = 0;
  trans.y = 0;
  trans.z = 0;
  numPoints = 0;
  numNormals = 0;
  numPolys = 0;
  point = new Point* [1];
  normal = new Point* [1];
  poly = new Polygon* [1];
}

Object::~Object()
{
  for (count = 0; count < numPolys; count++) delete poly[count];
  for (count = 0; count < numNormals; count++) delete normal[count];
  for (count = 0; count < numPoints; count++) delete point[count];
  delete poly;
  delete point;
}

void Object::SetFlipMatrix (float mat[9])
{
  trmat[0] = mat[0]; trmat[1] = mat[2]; trmat[2] = mat[1];
  trmat[3] = mat[6]; trmat[4] = mat[8]; trmat[5] = mat[7];
  trmat[6] = mat[3]; trmat[7] = mat[5]; trmat[8] = mat[4];
}

void Object::SetMatrix (float mat[9])
{
  trmat[0] = mat[0]; trmat[1] = mat[1]; trmat[2] = mat[2];
  trmat[3] = mat[3]; trmat[4] = mat[4]; trmat[5] = mat[5];
  trmat[6] = mat[6]; trmat[7] = mat[7]; trmat[8] = mat[8];
}

void Object::SetScale (Vector svec)
{
  if (scene->doscale) {
    if ((svec.x < 0.01)||(svec.y < 0.01)||(svec.z < 0.01)) {
      scal.x = 0.01;
      scal.y = 0.01;
      scal.z = 0.01;
    } else scal = svec;
  } else {
    scal.x = 1.0;
    scal.y = 1.0;
    scal.z = 1.0;
  }
}

void Object::SetPosition (Vector pos)
{
  trans = pos;
}

void Object::SetPosition (float x, float y, float z)
{
  trans.x = x;
  trans.y = y;
  trans.z = z;
}

void Object::SetPivot (Vector piv)
{
  pivot = piv;
}

void Object::SetShading (int surf, int shade, int flag, int c, char *tex, char *tex2)
{
  for (count = 0; count < numPolys; count++)
    poly[count]->SetShading (surf, shade, flag, c, tex, tex2);
}

void Object::AddPoint (float x, float y, float z)
{
  Point **temp;

  temp = new Point* [numPoints+1];
  memcpy (temp, point, numPoints * sizeof(Point*));
  delete point;
  point = temp;
  point[numPoints] = new Point (x, y, z);
  numPoints++;
}

void Object::AddPoint (Point *p)
{
  Point **temp;

  temp = new Point* [numPoints+1];
  memcpy(temp, point, numPoints * sizeof (Point*));
  delete point;
  point = temp;
  point[numPoints] = p;
  numPoints++;
}

void Object::AddNormal (float x, float y, float z)
{
  Point **temp;

  temp = new Point* [numNormals+1];
  memcpy (temp, normal, numNormals * sizeof (Point*));
  delete normal;
  normal = temp;
  normal[numNormals] = new Point (x, y, z);
  numNormals++;
}

void Object::AddNormal (Point *p)
{
  Point **temp;

  temp = new Point* [numNormals+1];
  memcpy(temp, normal, numNormals * sizeof (Point*));
  delete normal;
  normal = temp;
  normal[numNormals] = p;
  numNormals++;
}

void Object::AddPoly (Polygon *pg)
{
  Polygon **temp;

  temp = new Polygon*[numPolys + 1];
  memcpy (temp, poly, numPolys * sizeof (Polygon*));
  delete poly;
  poly = temp;
  poly[numPolys] = pg;
  numPolys++;
  AddNormal (pg->normal);
  scene->sortbuf[scene->numSort] = pg;
  scene->numSort++;
}

void Object::AddPoly (int p1, int p2, int p3)
{
  Polygon *pg = new Polygon (point[p1], point[p2], point[p3]);
  AddPoly (pg);
}

inline void TransformPoint (Point *p, float *matrix, Vector pivot, Vector pos, Vector sc)
{
  float a, b, c;

  a = p->localX*sc.x - pivot.x;
  b = p->localY*sc.y - pivot.y;
  c = p->localZ*sc.z - pivot.z;
  p->x3d = (a * matrix[0] + b * matrix[1] + c * matrix[2])+ pos.x;
  p->y3d = (a * matrix[3] + b * matrix[4] + c * matrix[5])+ pos.y;
  p->z3d = (a * matrix[6] + b * matrix[7] + c * matrix[8])+ pos.z;
}

void Object::TransformLocal ()
{
   Vector zero = {0.0, 0.0, 0.0};
   Vector bla =  {1.0, 1.0, 1.0};

   for (count = 0; count < numPoints; count++) {
     TransformPoint (point[count], trmat, pivot, trans, scal);
     TransformPoint (normal[count], trmat, zero, zero, bla);
   }
   for (count = 0; count < numPolys; count++)
     TransformPoint (poly[count]->normal, trmat, zero, zero, bla);
}

void Object::Save ()
{
  for (count = 0; count < numPoints; count++) point[count]->Save ();
  for (count = 0; count < numPoints; count++) normal[count]->Save ();
  for (count = 0; count < numPolys; count++) poly[count]->normal->Save ();
}

void Object::CalcVertNormals ()
{
  for (count = 0; count < numPolys; count++)
    poly[count]->CalcGNormals ();
  for (count = 0; count < numPoints; count++)
    point[count]->AverageNormal ();
  for (count = 0; count < numPoints; count++)
    if (point[count]->numNormals) {
      AddNormal (point[count]->normalX, point[count]->normalY,
		 point[count]->normalZ);
      point[count]->normal = normal[numNormals-1];
    }
}

void Object::EyeSpace ()
{
  float  tmpX, tmpY, tmpZ;
  Camera *camera = scene->camera;

  for (count = 0; count < numPoints; count++) {
    Point *p = point[count];
    tmpX = p->x3d - camera->position->x3d;
    tmpY = p->y3d - camera->position->y3d;
    tmpZ = p->z3d - camera->position->z3d;
    p->eyeX = (tmpX*camera->Ux) + (tmpY*camera->Uy) + (tmpZ*camera->Uz);
    p->eyeY = (tmpX*camera->Vx) + (tmpY*camera->Vy) + (tmpZ*camera->Vz);
    p->eyeZ = (tmpX*camera->Nx) + (tmpY*camera->Ny) + (tmpZ*camera->Nz);
    if (p->eyeZ > 0.0) {
      p->x2d = XPROJ + (int)((p->eyeX / p->eyeZ)*256.0);
      p->y2d = YPROJ - (int)((p->eyeY / p->eyeZ)*256.0);
    }
  }
}
