// mighty.engine version 0.10
// polygon class

#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

Polygon::Polygon ()
{
}

Polygon::Polygon (Point *p1, Point *p2, Point *p3)
{
  normal = new Point;
  SetTo (p1, p2, p3);
  CalcNormal ();
}

Polygon::~Polygon()
{
  delete p1;
  delete p2;
  delete p3;
  delete normal;
}

void Polygon::SetTo (Point *P1, Point *P2, Point *P3)
{
  p1 = P1;
  p2 = P2;
  p3 = P3;
}

void Polygon::CalcNormal()
{
  float a, b, c, magnitude;
  float d1[3], d2[3];

  d1[0] = p2->localX - p1->localX;
  d2[0] = p3->localX - p1->localX;
  d1[1] = p2->localY - p1->localY;
  d2[1] = p3->localY - p1->localY;
  d1[2] = p2->localZ - p1->localZ;
  d2[2] = p3->localZ - p1->localZ;
  a = d1[2]*d2[1] - d1[1]*d2[2];
  b = d1[0]*d2[2] - d1[2]*d2[0];
  c = d1[1]*d2[0] - d1[0]*d2[1];
  magnitude = SQRT ((a*a) + (b*b) + (c*c));
  a /= magnitude;
  b /= magnitude;
  c /= magnitude;
  normal->SetCoords (a, b, c);
}

void Polygon::SetShading (int surf, int shade, int flag, int c, char *tex, char *tex2)
{
  surface  = surf;
  shading  = shade;
  flags    = flag;
  color    = c;
  texture  = tex;
  shadetex = tex2;
}

void Polygon::CalcGNormals ()
{
  p1->AddNormal (normal->localX, normal->localY, normal->localZ);
  p2->AddNormal (normal->localX, normal->localY, normal->localZ);
  p3->AddNormal (normal->localX, normal->localY, normal->localZ);
}

void Polygon::Render (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
  if ((MIN (a->x2, MIN (b->x2, c->x2)) >= MAXX) ||
      (MIN (a->y2, MIN (b->y2, c->y2)) >= MAXY) ||
      (MAX (a->x2, MAX (b->x2, c->x2)) < 0)     ||
      (MAX (a->y2, MAX (b->y2, c->y2)) < 0)) return;

  switch (surface) {
    case sf_dot: polyDot (a, b, c); break;
    case sf_wireframe: polyWire (a, b, c); break;
    case sf_plain: switch (shading) {
      case sh_ambient:
      case sh_lambert: polyFlat (a, b, c); break;
      case sh_gouraud: polyGouraud (a, b, c); break;
      case sh_environment: polyTexture (a, b, c, texture);
    } break;
    case sf_texture: switch (shading) {
      case sh_ambient: polyTexture (a, b, c, texture); break;
    }
  }
}

void Polygon::ZRender (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
  a->x2 = XPROJ + (int)((a->x/a->z)*256.0);
  a->y2 = YPROJ - (int)((a->y/a->z)*256.0);
  b->x2 = XPROJ + (int)((b->x/b->z)*256.0);
  b->y2 = YPROJ - (int)((b->y/b->z)*256.0);
  c->x2 = XPROJ + (int)((c->x/c->z)*256.0);
  c->y2 = YPROJ - (int)((c->y/c->z)*256.0);
  if ((MIN (a->x2, MIN (b->x2, c->x2)) >= MAXX) ||
      (MIN (a->y2, MIN (b->y2, c->y2)) >= MAXY) ||
      (MAX (a->x2, MAX (b->x2, c->x2)) < 0)     ||
      (MAX (a->y2, MAX (b->y2, c->y2)) < 0)) return;
  switch (surface) {
    case sf_dot: polyDot (a, b, c); break;
    case sf_wireframe: polyWire (a, b, c); break;
    case sf_plain: switch (shading) {
      case sh_ambient:
      case sh_lambert: polyFlat (a, b, c); break;
      case sh_gouraud: polyGouraud (a, b, c); break;
      case sh_environment: polyTexture (a, b, c, texture);
    } break;
    case sf_texture: switch (shading) {
      case sh_ambient: polyTexture (a, b, c, texture); break;
    }
  }
}

void Polygon::Display ()
{
  float     dot;
  int       i1, i2, i3;
  int       u1, v1, u2, v2, u3, v3;
  int       numclip, clipcode = 0;
  ClipPoint src[3], dst[4];

  SubVec (scene->camera->position, p1);
  dot = DotProduct (normal, tPoint);
  if ((dot > 0.0) || (flags & pf_twosided)) {
    if (p1->eyeZ < CLIPDISTANCE) clipcode |= 1;
    if (p2->eyeZ < CLIPDISTANCE) clipcode |= 2;
    if (p3->eyeZ < CLIPDISTANCE) clipcode |= 4;
    if (clipcode == 7) return; // totally behind camera

    src[0].x2 = p1->x2d; src[0].y2 = p1->y2d;
    src[1].x2 = p2->x2d; src[1].y2 = p2->y2d;
    src[2].x2 = p3->x2d; src[2].y2 = p3->y2d;
    src[0].u1 = p1->u;   src[0].v1 = p1->v;
    src[1].u1 = p2->u;   src[1].v1 = p2->v;
    src[2].u1 = p3->u;   src[2].v1 = p3->v;

    switch (shading) {
      case sh_ambient: src[0].gc = color; break;
      case sh_lambert: src[0].gc = GouraudLight (p1, normal)+color; break;
      case sh_gouraud: src[0].c = GouraudLight (p1, p1->normal)+color;
                       src[1].c = GouraudLight (p2, p2->normal)+color;
                       src[2].c = GouraudLight (p3, p3->normal)+color;
                       break;
      case sh_environment: src[0].u1 = asine_table [int (p1->normal->x3d*127.0)+127];
                           src[0].v1 = asine_table [int (p1->normal->y3d*127.0)+127];
                           src[1].u1 = asine_table [int (p2->normal->x3d*127.0)+127];
                           src[1].v1 = asine_table [int (p2->normal->y3d*127.0)+127];
                           src[2].u1 = asine_table [int (p3->normal->x3d*127.0)+127];
                           src[2].v1 = asine_table [int (p3->normal->y3d*127.0)+127];
                           break;
    }
    if (clipcode == 0) Render (&src[0], &src[1], &src[2]);
    else {
      src[0].x = p1->eyeX; src[0].y = p1->eyeY; src[0].z = p1->eyeZ;
      src[1].x = p2->eyeX; src[1].y = p2->eyeY; src[1].z = p2->eyeZ;
      src[2].x = p3->eyeX; src[2].y = p3->eyeY; src[2].z = p3->eyeZ;
      numclip = ZClip (&src[0], &dst[0], clipcode);
      dst[0].gc = src[0].gc;
      ZRender (&dst[0], &dst[1], &dst[2]);
      if (numclip == 4) ZRender (&dst[0], &dst[2], &dst[3]);
    }
  }
}
