// mighty.engine version 0.10
// quaternion routines

#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

void qt_matrix (quat q, float *mat)
{
  double norm = q.x*q.x + q.y*q.y + q.z*q.z + q.w*q.w;
  double s = (norm > 0.0) ? 2.0/norm : 0.0;
  double xs = q.x*s,  ys = q.y*s,  zs = q.z*s;
  double wx = q.w*xs, wy = q.w*ys, wz = q.w*zs,
	 xx = q.x*xs, xy = q.x*ys, xz = q.x*zs,
	 yy = q.y*ys, yz = q.y*zs, zz = q.z*zs;

  mat[0] = 1.0 - (yy + zz); mat[1] = xy + wz; mat[2] = xz - wy;
  mat[3] = xy - wz; mat[4] = 1.0 - (xx + zz); mat[5] = yz + wx;
  mat[6] = xz + wy; mat[7] = yz - wx; mat[8] = 1.0 - (xx + yy);
}

void qt_invmatrix (quat q, float *mat)
{
  float s,  xs, ys, zs, wx, wy, zz;
  float wz, xx, xy, xz, yy, yz, den;

  den =  (q.x*q.x + q.y*q.y + q.z*q.z + q.w*q.w);
  if (den == 0.0) s = 1.0; else s = 2.0/den;

  xs = q.x * s;   ys = q.y * s;  zs = q.z * s;
  wx = q.w * xs;  wy = q.w * ys; wz = q.w * zs;
  xx = q.x * xs;  xy = q.x * ys; xz = q.x * zs;
  yy = q.y * ys;  yz = q.y * zs; zz = q.z * zs;

  mat[0] = 1.0 - (yy + zz); mat[1] = xy - wz; mat[2] = xz + wy;
  mat[3] = xy + wz; mat[4] = 1.0 - (xx + zz); mat[5] = yz - wx;
  mat[6] = xz - wy; mat[7] = yz + wx; mat[8] = 1.0 - (xx + yy);
}

void qt_mul (quat qL, quat qR, quat *qq)
{
  qq->w = qL.w*qR.w - qL.x*qR.x - qL.y*qR.y - qL.z*qR.z;
  qq->x = qL.w*qR.x + qL.x*qR.w + qL.y*qR.z - qL.z*qR.y;
  qq->y = qL.w*qR.y + qL.y*qR.w + qL.z*qR.x - qL.x*qR.z;
  qq->z = qL.w*qR.z + qL.z*qR.w + qL.x*qR.y - qL.y*qR.x;
}

void qt_convert (float w, float x, float y, float z, quat *q)
{
  q->w = cos (w/2);
  q->x = sin (w/2)*x;
  q->y = sin (w/2)*y;
  q->z = sin (w/2)*z;
}
