// mighty.engine version 0.10
// rendering routines
// this is just mri's txtdrwc.cpp texturemapper.

#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

static Vertex *left_array[3], *right_array[3];
static int     left_section, right_section;
static int     left_section_height, right_section_height;
static int     left_u, delta_left_u, left_v, delta_left_v;
static int     left_c, delta_left_c;
static int     left_x, delta_left_x, right_x, delta_right_x;
static int     vertclip, horzclip;

int shl10idiv (int x, int y);
#pragma aux shl10idiv =  \
  " mov   edx, eax "     \
  " shl   eax, 10  "     \
  " sar   edx, 22  "     \
  " idiv  ebx      "     \
  parm [eax] [ebx]       \
  modify exact [eax edx] \
  value [eax];

void polyDot (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
}

void polyWire (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
}

inline int tex_RightSection ()
{
  Vertex *v1 = right_array[right_section];
  Vertex *v2 = right_array[right_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_right_x = ((v2->x - v1->x) << 16) / height;
    right_x = v1->x << 16;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_right_x = ((v2->x - v1->x) << 16) / oheight;
    right_x = (v1->x << 16) + delta_right_x * y2;
  }
  right_section_height = height;
  return height;
}

inline static int tex_LeftSection ()
{
  Vertex *v1 = left_array[left_section];
  Vertex *v2 = left_array[left_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_left_x = ((v2->x - v1->x) << 16) / height;
    left_x = v1->x << 16;
    delta_left_u = ((v2->u - v1->u) << 10) / height;
    left_u = v1->u << 10;
    delta_left_v = ((v2->v - v1->v) << 10) / height;
    left_v = v1->v << 10;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_left_x = ((v2->x - v1->x) << 16) / oheight;
    left_x = (v1->x << 16) + delta_left_x * y2;
    delta_left_u = ((v2->u - v1->u) << 10) / oheight;
    left_u = (v1->u << 10) + delta_left_u * y2;
    delta_left_v = ((v2->v - v1->v) << 10) / oheight;
    left_v = (v1->v << 10) + delta_left_v * y2;
  }
  left_section_height = height;
  return height;
}

inline int gour_RightSection ()
{
  Vertex *v1 = right_array[right_section];
  Vertex *v2 = right_array[right_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_right_x = ((v2->x - v1->x) << 16) / height;
    right_x = v1->x << 16;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_right_x = ((v2->x - v1->x) << 16) / oheight;
    right_x = (v1->x << 16) + delta_right_x * y2;
  }
  right_section_height = height;
  return height;
}

inline static int gour_LeftSection ()
{
  Vertex *v1 = left_array[left_section];
  Vertex *v2 = left_array[left_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_left_x = ((v2->x - v1->x) << 16) / height;
    left_x = v1->x << 16;
    delta_left_c = ((v2->c - v1->c) << 10) / height;
    left_c = v1->c << 10;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_left_x = ((v2->x - v1->x) << 16) / oheight;
    left_x = (v1->x << 16) + delta_left_x * y2;
    delta_left_c = ((v2->c - v1->c) << 10) / oheight;
    left_c = (v1->c << 10) + delta_left_c * y2;
  }
  left_section_height = height;
  return height;
}

inline int flat_RightSection ()
{
  Vertex *v1 = right_array[right_section];
  Vertex *v2 = right_array[right_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_right_x = ((v2->x - v1->x) << 16) / height;
    right_x = v1->x << 16;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_right_x = ((v2->x - v1->x) << 16) / oheight;
    right_x = (v1->x << 16) + delta_right_x * y2;
  }
  right_section_height = height;
  return height;
}

inline static int flat_LeftSection ()
{
  Vertex *v1 = left_array[left_section];
  Vertex *v2 = left_array[left_section-1];
  int height;

  if (!vertclip) {
    height = v2->y - v1->y;
    if (height == 0) return 0;
    delta_left_x = ((v2->x - v1->x) << 16) / height;
    left_x = v1->x << 16;
  } else {
    int y1 = v1->y;
    int y2 = v2->y;
    int oheight = height = y2-y1;
    if ((y2 -= CLIPY) > 0) height -= y2;
    y2 = 0;
    if (y1 < 0) y2 = 0 - y1;
    if ((height -= y2) <= 0) return height;
    delta_left_x = ((v2->x - v1->x) << 16) / oheight;
    left_x = (v1->x << 16) + delta_left_x * y2;
  }
  left_section_height = height;
  return height;
}

void polyFlat (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
  Vertex v1, v2, v3;
  int top;
  int col=a->gc;

  v1.x=a->x2; v1.y=a->y2;
  v2.x=b->x2; v2.y=b->y2;
  v3.x=c->x2; v3.y=c->y2;
  if (v1.y > v2.y) { Vertex v=v1; v1=v2; v2=v; }
  if (v1.y > v3.y) { Vertex v=v1; v1=v3; v3=v; }
  if (v2.y > v3.y) { Vertex v=v2; v2=v3; v3=v; }

  if ((top = v1.y) < 0) {
    top = 0;
    vertclip = 1;
  } else if (v3.y > CLIPY) vertclip = 1;
    else vertclip = 0;
  int height = v3.y - v1.y;
  if (height == 0) return;
  int temp = ((v2.y - v1.y)<<16)/height;
  int longest = temp*(v3.x - v1.x)+((v1.x - v2.x)<<16);
  if (longest == 0) return;
  if (longest < 0) {
    right_array[0] = &v3;
    right_array[1] = &v2;
    right_array[2] = &v1;
    right_section  = 2;
    left_array[0]  = &v3;
    left_array[1]  = &v1;
    left_section   = 1;
    if (flat_LeftSection () <= 0) return;
    if (flat_RightSection () <= 0) {
      right_section--;
      if (flat_RightSection () <= 0) return;
    }
    if (longest > -0x1000) longest = -0x1000;
  } else {
    left_array[0]  = &v3;
    left_array[1]  = &v2;
    left_array[2]  = &v1;
    left_section   = 2;
    right_array[0] = &v3;
    right_array[1] = &v1;
    right_section  = 1;
    if (flat_RightSection () <= 0) return;
    if (flat_LeftSection () <= 0) {
      left_section--;
      if (flat_LeftSection () <= 0) return;
    }
    if (longest < 0x1000) longest = 0x1000;
  }
  if ((v1.x > CLIPX) || (v1.x < 0) || (v2.x > CLIPX) ||
      (v2.x < 0) || (v3.x > CLIPX) || (v3.x < 0)) horzclip = 1;
    else horzclip = 0;
  char *destptr = (char *)(top*MAXX + VirtualScreen);
  for (;;) {
    int x1, x2, width;
    if (!horzclip) {
      x1 = left_x >> 16;
      width = (right_x >> 16) - x1;
    } else {
      x1 = left_x >> 16;
      x2 = right_x >> 16;
      width = x2-x1;
      if ((x2 -= CLIPX) > 0) width -= x2;
      x2 = 0;
      if (x1 < 0) {
        x2 = 0 - x1;
        x1 = 0;
      }
      if ((width -= x2) <= 0) goto linedone;
    }
    if (width > 0) {
      char *dest = destptr + x1;
      do { *dest++ = col; } while (--width);
    }
linedone:
    destptr += MAXX;
    if (--left_section_height == 0) {
      if (--left_section <= 0) return;
      if (flat_LeftSection () <= 0) return;
    } else {
      left_x += delta_left_x;
    }
    if (--right_section_height == 0) {
      if (--right_section <= 0) return;
      if (flat_RightSection () <= 0) return;
    } else right_x += delta_right_x;
  }
}

void polyGouraud (ClipPoint *a, ClipPoint *b, ClipPoint *c)
{
  Vertex v1, v2, v3;
  int top;

  v1.x=a->x2; v1.y=a->y2; v1.c=a->c<<6;
  v2.x=b->x2; v2.y=b->y2; v2.c=b->c<<6;
  v3.x=c->x2; v3.y=c->y2; v3.c=c->c<<6;
  if (v1.y > v2.y) { Vertex v=v1; v1=v2; v2=v; }
  if (v1.y > v3.y) { Vertex v=v1; v1=v3; v3=v; }
  if (v2.y > v3.y) { Vertex v=v2; v2=v3; v3=v; }

  if ((top = v1.y) < 0) {
    top = 0;
    vertclip = 1;
  } else if (v3.y > CLIPY) vertclip = 1;
    else vertclip = 0;
  int height = v3.y - v1.y;
  if (height == 0) return;
  int temp = ((v2.y - v1.y)<<16)/height;
  int longest = temp*(v3.x - v1.x)+((v1.x - v2.x)<<16);
  if (longest == 0) return;
  if (longest < 0) {
    right_array[0] = &v3;
    right_array[1] = &v2;
    right_array[2] = &v1;
    right_section  = 2;
    left_array[0]  = &v3;
    left_array[1]  = &v1;
    left_section   = 1;
    if (gour_LeftSection () <= 0) return;
    if (gour_RightSection () <= 0) {
      right_section--;
      if (gour_RightSection () <= 0) return;
    }
    if (longest > -0x1000) longest = -0x1000;
  } else {
    left_array[0]  = &v3;
    left_array[1]  = &v2;
    left_array[2]  = &v1;
    left_section   = 2;
    right_array[0] = &v3;
    right_array[1] = &v1;
    right_section  = 1;
    if (gour_RightSection () <= 0) return;
    if (gour_LeftSection () <= 0) {
      left_section--;
      if (gour_LeftSection () <= 0) return;
    }
    if (longest < 0x1000) longest = 0x1000;
  }
  if ((v1.x > CLIPX) || (v1.x < 0) || (v2.x > CLIPX) ||
      (v2.x < 0) || (v3.x > CLIPX) || (v3.x < 0)) horzclip = 1;
    else horzclip = 0;
  int dcdx = shl10idiv (temp*(v3.c - v1.c)+((v1.c - v2.c)<<16), longest);
  int dzdx = shl10idiv (temp*(v3.z - v1.z)+((v1.z - v2.z)<<16), longest);
  char *destptr = &VirtualScreen[top*MAXX];
  for (;;) {
    int x1, x2, width, c, z;
    if (!horzclip) {
      x1 = left_x >> 16;
      width = (right_x >> 16) - x1;
      c = left_c;
    } else {
      x1 = left_x >> 16;
      x2 = right_x >> 16;
      width = x2-x1;
      if ((x2 -= CLIPX) > 0) width -= x2;
      x2 = 0;
      if (x1 < 0) {
        x2 = 0 - x1;
        x1 = 0;
      }
      if ((width -= x2) <= 0) goto linedone;
      c = (left_c + dcdx * x2);
    }
    if (width > 0) {
      char *dest = destptr + x1;
      int dc = dcdx;
      int dz = dzdx;
      do {
        *dest = c>>16;
        c += dc;
        dest++;
      } while (--width);
    }
linedone:
    destptr += MAXX;
    if (--left_section_height == 0) {
      if (--left_section <= 0) return;
      if (gour_LeftSection () <= 0) return;
    } else {
      left_x += delta_left_x;
      left_c += delta_left_c;
    }
    if (--right_section_height == 0) {
      if (--right_section <= 0) return;
      if (gour_RightSection () <= 0) return;
    } else right_x += delta_right_x;
  }
}

void polyTexture (ClipPoint *a, ClipPoint *b, ClipPoint *c, char *bitmap)
{
  Vertex v1, v2, v3;
  int top;

  v1.x=a->x2;v1.y=a->y2;v1.u=a->u1;v1.v=a->v1;
  v2.x=b->x2;v2.y=b->y2;v2.u=b->u1;v2.v=b->v1;
  v3.x=c->x2;v3.y=c->y2;v3.u=c->u1;v3.v=c->v1;
  if (v1.y > v2.y) { Vertex v=v1; v1=v2; v2=v; }
  if (v1.y > v3.y) { Vertex v=v1; v1=v3; v3=v; }
  if (v2.y > v3.y) { Vertex v=v2; v2=v3; v3=v; }

  if ((top = v1.y) < 0) {
    top = 0;
    vertclip = 1;
  } else if (v3.y > CLIPY) vertclip = 1;
    else vertclip = 0;
  int height = v3.y - v1.y;
  if (height == 0) return;
  int temp = ((v2.y - v1.y)<<16)/height;
  int longest = temp*(v3.x - v1.x)+((v1.x - v2.x)<<16);
  if (longest == 0) return;
  if (longest < 0) {
    right_array[0] = &v3;
    right_array[1] = &v2;
    right_array[2] = &v1;
    right_section  = 2;
    left_array[0]  = &v3;
    left_array[1]  = &v1;
    left_section   = 1;
    if (tex_LeftSection () <= 0) return;
    if (tex_RightSection () <= 0) {
      right_section--;
      if (tex_RightSection () <= 0) return;
    }
    if (longest > -0x1000) longest = -0x1000;
  } else {
    left_array[0]  = &v3;
    left_array[1]  = &v2;
    left_array[2]  = &v1;
    left_section   = 2;
    right_array[0] = &v3;
    right_array[1] = &v1;
    right_section  = 1;
    if (tex_RightSection () <= 0) return;
    if (tex_LeftSection () <= 0) {
      left_section--;
      if (tex_LeftSection () <= 0) return;
    }
    if (longest < 0x1000) longest = 0x1000;
  }
  if ((v1.x > CLIPX) || (v1.x < 0) || (v2.x > CLIPX) ||
      (v2.x < 0) || (v3.x > CLIPX) || (v3.x < 0)) horzclip = 1;
    else horzclip = 0;
  int dudx = shl10idiv (temp*(v3.u - v1.u)+((v1.u - v2.u)<<16), longest);
  int dvdx = shl10idiv (temp*(v3.v - v1.v)+((v1.v - v2.v)<<16), longest);
  char *destptr = (char *)(top*MAXX + VirtualScreen);
  for (;;) {
    int x1, x2, width, u, v;
    if (!horzclip) {
      x1 = left_x >> 16;
      width = (right_x >> 16) - x1;
      u = left_u;
      v = left_v;
    } else {
      x1 = left_x >> 16;
      x2 = right_x >> 16;
      width = x2-x1;
      if ((x2 -= CLIPX) > 0) width -= x2;
      x2 = 0;
      if (x1 < 0) {
        x2 = 0 - x1;
        x1 = 0;
      }
      if ((width -= x2) <= 0) goto linedone;
      u = (left_u + dudx * x2);
      v = (left_v + dvdx * x2);
    }
    if (width > 0) {
      char *dest = destptr + x1;
      int du = dudx;
      int dv = dvdx;
      do {
        *dest++ = bitmap [((v>>8) & 0xff00) + ((u>>16) & 0x00ff)];
        u += du;
        v += dv;
      } while (--width);
    }
linedone:
    destptr += MAXX;
    if (--left_section_height == 0) {
      if (--left_section <= 0) return;
      if (tex_LeftSection () <= 0) return;
    } else {
      left_x += delta_left_x;
      left_u += delta_left_u;
      left_v += delta_left_v;
    }
    if (--right_section_height == 0) {
      if (--right_section <= 0) return;
      if (tex_RightSection () <= 0) return;
    } else right_x += delta_right_x;
  }
}
