// mighty.engine version 0.10
// scene class

#include <string.h>
#include <stdlib.h>
#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

Scene::Scene ()
{
  numObjects = 0;
  numLights = 0;
  numCameras = 0;
  numSort = 0;
  doscale = 1;
  object = new Object*[1];
  light = new Point*[1];
  cameras = new Camera*[1];
  sortbuf = new Polygon*[8192];
  keyframer = NULL;
}

Scene::~Scene ()
{
  int count;

  for (count = 0; count < numObjects; count++) delete object[count];
  for (count = 0; count < numLights; count++) delete light[count];
  for (count = 0; count < numCameras; count++) delete cameras[count];
  delete object;
  delete light;
  delete cameras;
  delete sortbuf;
  if (keyframer) delete keyframer;
}

void Scene::AddObject ()
{
  Object **temp;

  temp = new Object*[numObjects+1];
  memcpy (temp, object, numObjects*sizeof (Object*));
  delete object;
  object = temp;
  object[numObjects] = new Object;
  numObjects++;
}

void Scene::AddLight (float x, float y, float z)
{
  Point **temp;

  temp = new Point*[numLights+1];
  memcpy (temp, light, numLights * sizeof (Point*));
  delete light;
  light = temp;
  light[numLights] = new Point (x, y, z);
  numLights++;
}

void Scene::AddCamera (float vx, float vy, float vz,
		       float tx, float ty, float tz)
{
  Camera **temp;

  temp = new Camera*[numCameras+1];
  memcpy(temp, cameras, numCameras*sizeof (Camera*));
  delete cameras;
  cameras = temp;
  cameras[numCameras] = new Camera (vx, vy, vz, tx, ty, tz);
  numCameras++;
}

void Scene::SetActive ()
{
  scene = this;
}

void Scene::EyeSpace ()
{
  int i;
  for (i = 0; i < numObjects; i++) object[i]->EyeSpace ();
}

void Scene::Display ()
{
  int i;

  sortbuffer = sortbuf;
  switch (sorttype) {
    case s_zmin: for (i = 0; i < numSort; i++)
      sortbuf[i]->midZ = MIN (sortbuf[i]->p1->eyeZ,
                         MIN (sortbuf[i]->p2->eyeZ, sortbuf[i]->p3->eyeZ));
      break;
    case s_zmax: for (i = 0; i < numSort; i++)
      sortbuf[i]->midZ = MAX (sortbuf[i]->p1->eyeZ,
                         MAX (sortbuf[i]->p2->eyeZ, sortbuf[i]->p3->eyeZ));
      break;
    case s_zmid: for (i = 0; i < numSort; i++)
      sortbuf[i]->midZ = (sortbuf[i]->p1->eyeZ+
                          sortbuf[i]->p2->eyeZ+ sortbuf[i]->p3->eyeZ);
  }
  quickSort (numSort);
  for (i = 0; i < numSort; i++) sortbuf[i]->Display ();
}
