// mighty.engine version 0.10
// spline routines, by Adept/Esteem

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "engine.h"

//////////////////////////////////////////////////////////////////////////////

static Key keys[MAXKEYS];
static int numKeys;

float Ease (float t, float a, float b)
{
  float k;
  float s = a+b;

  if (s == 0.0) return t;
  if (s > 1.0) {
    a = a/s;
    b = b/s;
  }
  k = 1.0/(2.0-a-b);
  if (t < a) return ((k/a)*t*t);
    else {
      if (t < 1.0-b) return (k*(2*t - a));
        else {
          t = 1.0-t;
          return (1.0-(k/b)*t*t);
        }
    }
}

static void CompElementDeriv (float pp, float p, float pn,
                              float *ds, float *dd, float ksm,
                              float ksp, float kdm, float kdp)
{
  float delm, delp;

  delm = p - pp;
  delp = pn - p;
  *ds  = ksm*delm + ksp*delp;
  *dd  = kdm*delm + kdp*delp;
}

void CompDeriv (Key *keyp, Key *key, Key *keyn)
{
  float tm, cm, cp, bm, bp, tmcm, tmcp;
  float ksm, ksp, kdm, kdp, delm, delp;
  float dt, fp, fn, c;

  dt = .5 * (keyn->frame - keyp->frame);
  fp = (key->frame-keyp->frame)/dt;
  fn = (keyn->frame-key->frame)/dt;
  c  = fabs (key->cont);
  fp = fp+c-c*fp;
  fn = fn+c-c*fn;
  cm = 1.0-key->cont;
  tm = 0.5*(1.0-key->tens);
  cp = 2.0-cm;
  bm = 1.0-key->bias;
  bp = 2.0-bm;
  tmcm = tm*cm;
  tmcp = tm*cp;
  ksm = tmcm*bp*fp; ksp = tmcp*bm*fp;
  kdm = tmcp*bp*fn; kdp = tmcm*bm*fn;
  CompElementDeriv (keyp->pos.x, key->pos.x, keyn->pos.x,
                    &key->ds.x, &key->dd.x, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->pos.y, key->pos.y, keyn->pos.y,
                    &key->ds.y, &key->dd.y, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->pos.z, key->pos.z, keyn->pos.z,
                    &key->ds.z, &key->dd.z, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->pos.w, key->pos.w, keyn->pos.w,
                    &key->ds.w, &key->dd.w, ksm, ksp, kdm, kdp);
}

void CompDerivFirst (Key *key, Key *keyn, Key *keynn)
{
  float f20, f10, v20, v10;

  f20 = keynn->frame-key->frame;
  f10 = keyn->frame-key->frame;
  v20 = keynn->pos.x-key->pos.x;
  v10 = keyn->pos.x-key->pos.x;
  key->dd.x = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->pos.y-key->pos.y;
  v10 = keyn->pos.y-key->pos.y;
  key->dd.y = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->pos.z-key->pos.z;
  v10 = keyn->pos.z-key->pos.z;
  key->dd.z = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->pos.w-key->pos.w;
  v10 = keyn->pos.w-key->pos.w;
  key->dd.w = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
}

void CompDerivLast (Key *keypp, Key *keyp, Key *key)
{
  float f20, f10, v20, v10;

  f20 = key->frame-keypp->frame;
  f10 = key->frame-keyp->frame;
  v20 = key->pos.x-keypp->pos.x;
  v10 = key->pos.x-keyp->pos.x;
  key->ds.x = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->pos.y-keypp->pos.y;
  v10 = key->pos.y-keyp->pos.y;
  key->ds.y = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->pos.z-keypp->pos.z;
  v10 = key->pos.z-keyp->pos.z;
  key->ds.z = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->pos.w-keypp->pos.w;
  v10 = key->pos.w-keyp->pos.w;
  key->ds.w = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
}

void SetKey (Key *key, float frame, float x, float y, float z, float w,
             float t, float c, float b, float eto, float efrom)
{
  key->frame = frame;
  key->pos.x = x;
  key->pos.y = y;
  key->pos.z = z;
  key->pos.w = w;
  key->tens = t;
  key->cont = c;
  key->bias = b;
  key->easeto = eto;
  key->easefrom = efrom;
  key->ds.x = 0; key->ds.y = 0; key->ds.z = 0; key->ds.w = 0;
  key->dd.x = 0; key->dd.y = 0; key->dd.z = 0; key->dd.w = 0;
}

void InitKeys ()
{
  numKeys=0;
}

void AddKey (float frame, float x, float y, float z, float w,
             float t, float c, float b, float eto, float efrom)
{
  SetKey (&keys[numKeys], frame, x, y, z, w, t, c, b, eto, efrom);
  numKeys++;
}

void DrawSpline3d (Vector *frame)
{
  int   i, n, count = 0;
  float j, t, t2, t3, frames;
  float h[4];

  if (numKeys < 3) {
    for (n=0; n<scene->keyframer->numFrames; n++) {
      frame[n].x = keys[0].pos.x;
      frame[n].y = keys[0].pos.y;
      frame[n].z = keys[0].pos.z;
    }
    return;
  }
  if (numKeys > 2) for (n=1; n<numKeys-1; n++)
    CompDeriv (&keys[n-1], &keys[n], &keys[n+1]);
  CompDerivFirst (&keys[0], &keys[1], &keys[2]);
  CompDerivLast (&keys[numKeys-3], &keys[numKeys-2], &keys[numKeys-1]);
  for (n=0; n<numKeys-1; n++)
  {
    frames = keys[n+1].frame - keys[n].frame;
    for (j = 0; j < frames; j++)
    {
      t = j/frames;
      t = Ease (t, keys[n].easefrom, keys[n+1].easeto);
      t2 = t*t;                                                                                         // t2 = t^2;
      t3 = t2*t;                                                                                        // t3 = t^3;
      h[0] = 2*t3-3*t2+1;
      h[1] = -2*t3+3*t2;
      h[2] = t3-2*t2+t;
      h[3] = t3-t2;
      frame[count].x = (h[0]*keys[n].pos.x)+(h[1]*keys[n+1].pos.x) +
                       (h[2]*keys[n].dd.x)+ (h[3]*keys[n+1].ds.x);
      frame[count].y = (h[0]*keys[n].pos.y)+(h[1]*keys[n+1].pos.y) +
                       (h[2]*keys[n].dd.y)+ (h[3]*keys[n+1].ds.y);
      frame[count].z = (h[0]*keys[n].pos.z)+(h[1]*keys[n+1].pos.z) +
                       (h[2]*keys[n].dd.z)+ (h[3]*keys[n+1].ds.z);
      if ((count++)>=scene->keyframer->numFrames) return;
    }
  }
}

void DrawSpline4d (quat *frame)
{
  int   i, n, count = 0;
  float j, t, t2, t3, frames;
  float h[4];

  if (numKeys < 3) {
    for (n=0; n<scene->keyframer->numFrames; n++) {
      frame[n].x = keys[0].pos.x;
      frame[n].y = keys[0].pos.y;
      frame[n].z = keys[0].pos.z;
      frame[n].w = keys[0].pos.w;
    }
    return;
  }
  if (numKeys > 2) for (n=1; n<numKeys-1; n++)
    CompDeriv (&keys[n-1], &keys[n], &keys[n+1]);
  CompDerivFirst (&keys[0], &keys[1], &keys[2]);
  CompDerivLast (&keys[numKeys-3], &keys[numKeys-2], &keys[numKeys-1]);
  for (n=0; n<numKeys-1; n++)
  {
    frames = keys[n+1].frame - keys[n].frame;
    for (j = 0; j < frames; j++)
    {
      t = j/frames;
      t = Ease (t, keys[n].easefrom, keys[n+1].easeto);
      t2 = t*t;                                                                                         // t2 = t^2;
      t3 = t2*t;                                                                                        // t3 = t^3;
      h[0] = 2*t3-3*t2+1;
      h[1] = -2*t3+3*t2;
      h[2] = t3-2*t2+t;
      h[3] = t3-t2;
      frame[count].x = (h[0]*keys[n].pos.x)+(h[1]*keys[n+1].pos.x) +
                       (h[2]*keys[n].dd.x)+ (h[3]*keys[n+1].ds.x);
      frame[count].y = (h[0]*keys[n].pos.y)+(h[1]*keys[n+1].pos.y) +
                       (h[2]*keys[n].dd.y)+ (h[3]*keys[n+1].ds.y);
      frame[count].z = (h[0]*keys[n].pos.z)+(h[1]*keys[n+1].pos.z) +
                       (h[2]*keys[n].dd.z)+ (h[3]*keys[n+1].ds.z);
      frame[count].w = (h[0]*keys[n].pos.w)+(h[1]*keys[n+1].pos.w) +
                       (h[2]*keys[n].dd.w)+ (h[3]*keys[n+1].ds.w);
      if ((count++)>=scene->keyframer->numFrames) return;
    }
  }
}
