(* VGA Library ------------------------------------------------------------ *)
(*                                                                          *)
(* MODULE NAME: Base Module                                                 *)
(* VERSION    : 0.10                                                        *)
(* LAST CHANGE: 13-07-95                                                    *)
(* AUTHOR     : BoRZoM / [TATOO]                                            *)
(* DIRECTIVES : *) {$G+,A+,E+,F+,R-,S-,D-} (*                               *)
(*                                                                          *)
(*                                                                          *)
(* NOTES      : There are several mode types supported by this unit:        *)
(*              standard and non-standard 100% register compatible vga      *)
(*              modes, they include standard 320x200x256,                   *)
(*              mode-x (unchained), mode-q (chained),                       *)
(*              mode-y (unchained, RGB page flipping) and fake RGB modes,   *)
(*              640x400x16 ($12) and textmode 512-char font table graphics. *)
(*                                                                          *)
(*              modex by matt prichard                                      *)
(*                                                                          *)
(*                          *) UNIT VGA;    (*                              *)

(****************************************************************************)
(******************************) INTERFACE (*********************************)
(****************************************************************************)

const                                                       { Mode constants }
  t80x25      = 0;    { text }
  t80x28      = 1;    { text }
  t80x50      = 2;    { text }
  tg80x25     = 3;    { text graphics }
  tg80x50     = 4;    { text graphics }
  g320x200x16 = 5;    { 16 color }
  g640x480x16 = 6;    { 16 color }
  g320x240x16 = 1684; { 16 color tweaked }
  g320x200    = 7;    { chained }
  g320x100    = 3393; { tweaked, chained }
  g320x80     = 4417; { tweaked, chained }
  g256x256    = 8;    { tweaked, chained }
  g256x128    = 9;    { tweaked, chained }
  g160x100    = 3345; { tweaked, chained }
  g160x80     = 4369; { tweaked, chained }
  g80x200     = 10;   { tweaked, chained }
  g80x80      = 11;   { tweaked, chained }
  g80x50      = 12;   { tweaked, chained }
  y320x200    = 512;  { tweaked, unchained, RGB }
  y320x400    = 513;  { tweaked, unchained, RGB }
  x320x200    = 255;  { tweaked, unchained }
  x320x400    = 256;  { tweaked, unchained }
  x360x200    = 257;  { tweaked, unchained }
  x360x400    = 258;  { tweaked, unchained }
  x320x240    = 259;  { tweaked, unchained }
  x320x480    = 260;  { tweaked, unchained }
  x360x240    = 261;  { tweaked, unchained }
  x360x480    = 262;  { tweaked, unchained }

  LogMaxX         : Word = 0;
  LogMaxY         : Word = 0;
  NumPages        : Byte = 0;
  CurMode         : Byte = 0;
  Current_Page    : Word = 0;
  Current_Segment : Word = 0;

  MinX : Word = 0;
  MinY : Word = 0;
  MaxX : Word = 320;
  MaxY : Word = 200;

  Gamma       : Array [0..63] of Byte =
  ( 0, 10, 14, 17, 19, 21, 23, 24, 26, 27, 28, 29, 31, 32, 33, 34,
   35, 36, 37, 37, 38, 39, 40, 41, 41, 42, 43, 44, 44, 45, 46, 46,
   47, 48, 48, 49, 49, 50, 51, 51, 52, 52, 53, 53, 54, 54, 55, 55,
   56, 56, 57, 57, 58, 58, 59, 59, 60, 60, 61, 61, 62, 62, 63, 63);
  TextPalette : Array [0..15] of Byte =
    (0, 1, 2, 3, 4, 5, 20, 7, 56, 57, 58, 59, 60, 61, 62, 63);
  CharOffsets : Array [0..7] of Word =
    ($0000, $4000, $8000, $C000, $2000, $6000, $A000, $E000);
  AsciiPix    : Array [0..15] of Char = ' -+oO$&#@';

Type
  Point3D  = Record
    x, y, z : Integer;
  End;
  Point2D  = Record
    x, y    : Integer;
  End;
  PIntArray = ^TIntArray;
  TIntArray = Array [0..0] of Integer;
  PByteArray = ^TByteArray;
  TByteArray = Array [0..0] of Byte;
  PLongArray = ^TLongArray;
  TLongArray = Array [0..0] of Longint;
  Morph    = Object
    v1, v2, Result : PIntArray;
    Stages, Size   : Word;

    Constructor Init (_Stages, _Size : Word);
    Procedure   Process (Stage : Word);
    Destructor  Done;
  End;

  BMorph    = Object
    v1, v2, Result : PByteArray;
    Stages, Size   : Word;

    Constructor Init (_Stages, _Size : Word);
    Procedure   Process (Stage : Word);
    Destructor  Done;
  End;

  LMorph    = Object
    v1, v2, Result : PLongArray;
    Stages, Size   : Word;

    Constructor Init (_Stages, _Size : Word);
    Procedure   Process (Stage : Word);
    Destructor  Done;
  End;

var
  Screen     : Array [0..65534] of Byte absolute $A000:$0000;
  tgCount    : Integer;
  PageSize   : Word;
  FadeStep   : Byte;
  vgaError   : Integer;

(** HARDWARE ****************************************************************)

Function  isVGA : Boolean;
Procedure SetMode (mode, x_vX, x_vY, x_p : Word);
Procedure WaitSync;
Procedure Set400Lines;
Procedure LineRepeat (b : Byte);
Procedure DisplayWidth (b : Byte);
Procedure WriteMode (b : Byte);
Procedure SetCharMap (b1, b2 : Byte);
Procedure BitPlane (b : Byte);
Procedure WriteMap (b : Byte);
Procedure PelPan (b : Byte);
Procedure SetAddress (w : Word);
Procedure SetSplit (w : Word);
Procedure SetCharWidth (b : Byte);
Procedure Unchain;
Procedure Rechain;
Procedure Overscan;
Procedure ScreenON;
Procedure ScreenOFF;
Procedure GetFont (var Buf);
Procedure SetFont (var Buf);
Procedure SetOneChar (var Buf; C : Byte);
Procedure SetCharSet (var Buf1, Buf2; Map1, Map2 : Integer);
Procedure IceColor (b : Boolean);
Procedure SetBorderColor (b : Byte);
Procedure SetCursor (Start, Stop : Byte);
Procedure FastWrite (x, y : Word; a : Byte; s : String);
Procedure Raster (var Buf; Col, Num : Integer);
Procedure BiosPix (x, y : Integer; col : Byte);
Procedure InitVRServer (routine : Pointer);
Procedure RestoreVRServer;

(** VIRTUAL SCREEN **********************************************************)

Procedure ClearPage (p : Pointer);
Procedure ClearPageColor (p : Pointer; c : Byte);
Procedure ClearPagePart (p : Pointer; x1, y1, x2, y2 : Integer);
Procedure ClearPagePartColor (p : Pointer; x1, y1, x2, y2, c : Integer);
Procedure ViewPage (p : Pointer);
Procedure ViewPagePart (p : Pointer; x1, y1, x2, y2 : Integer);

(** DRAW ROUTINES (any 256 color chained mode) ******************************)

Procedure Line (x1, y1, x2, y2 : Integer; c : Byte; Page : Word);
Procedure Ellipse (x, y, Rad : Integer; c : Byte; Page : Word);

(** SPECIAL MODE DRAW ROUTINES **********************************************)

Procedure g16Clear;
Procedure g16Pixel (x, y : Integer; c : Byte);
Procedure g16FillBox (x1, y1, x2, y2 : Integer; c : Byte);
Procedure g16hLine (x, y, d : Integer; c : Byte);
Procedure g16vLine (x, y, d : Integer; c : Byte);
Procedure g16Line (x1, y1, x2, y2 : Integer; c : Byte);
Procedure g16Ellipse (x, y, r : Integer; c : Byte);

(** PALETTE *****************************************************************)

Procedure SetRGB (c, r, g, b : Byte);
Procedure GetPalette (var Buf);
Procedure SetPalette (var Buf);
procedure BlankPalette;
Procedure InitFade;
Function  Fadeout : Boolean;
Procedure InitFadeTo (var Buf);
Function  FadeTo : Boolean;
Procedure RampPalette (var Buf; r1, g1, b1, r2, g2, b2, c1, c2 : Byte);
Procedure NegatePalette (var Buf; c1, c2 : Byte);
Procedure ScalePalette (var Buf; r, g, b, c1, c2 : Byte);
Procedure SlidePalette (var Buf; c1, c2 : Byte; Direction : Boolean);

(** MODE-X ROUTINES *********************************************************)

Procedure Clear_VGA_Screen (Color : Integer);
Procedure Set_Point (X, Y, Color : Integer);
Function  Read_Point (X, Y : Integer) : Integer;
Procedure Fill_Block (X1, Y1, X2, Y2, Color : Integer);
Procedure Draw_Line (X1, Y1, X2, Y2, Color : Integer);
Procedure Set_Active_Page (PageNo : Integer);
Function  Get_Active_Page : Integer;
Procedure Set_Display_Page (PageNo : Integer);
Function  Get_Display_Page : Integer;
Procedure Set_Window (PageNo, XOfs, YOfs : Integer);
Function  Get_X_Offset : Integer;
Function  Get_Y_Offset : Integer;
Procedure GPrintc (CNum, X, Y, ColorF, ColorB : Integer);
Procedure TGPrintc (CNum, X, Y, ColorF : Integer);
Procedure Print_Str (var Text; Len, X, Y,ColorF, ColorB : Integer);
Procedure TPrint_Str (var Text; Len, X, Y, ColorF : Integer);
Procedure Set_Display_Font (Var FontData; FontNum : Integer);
Procedure Draw_Bitmap (var Image; X, Y, W, H : Integer);
Procedure TDraw_Bitmap (var Image; X, Y, W, H : Integer);
Procedure Copy_Page (SourcePage, DestPage : Integer);
Procedure Copy_Bitmap (SourcePage, X1, Y1, X2, Y2,
          DestPage, DestX1, DestY1 : Integer);

(** MISCELLANEOUS ***********************************************************)

Procedure SetTimerFreq (w : Word);
Function  LoadCEL (fn : String) : Pointer;
Function  iSqrt (x : Longint) : Word;
Function  KeyHit : Boolean;

(****************************************************************************)
(****************************) IMPLEMENTATION (******************************)
(****************************************************************************)

{$L MODEX}
{$L VGALIB}
Procedure InitVRServer (routine : Pointer); External;
Procedure RestoreVRServer; External;
Function  Set_VGA_ModeX (Mode, MaxX, MaxY, Pages : Integer) : Integer; External;
Procedure Clear_VGA_Screen (Color : Integer);                  External;
Procedure Set_Point (X, Y, Color : Integer);                   External;
Function  Read_Point (X, Y : Integer) : Integer;               External;
Procedure Fill_Block (X1, Y1, X2, Y2, Color : Integer);        External;
Procedure Draw_Line (X1, Y1, X2, Y2, Color : Integer);         External;
Procedure Set_Active_Page (PageNo : Integer);                  External;
Function  Get_Active_Page : Integer;                           External;
Procedure Set_Display_Page (PageNo : Integer);                 External;
Function  Get_Display_Page : Integer;                          External;
Procedure Set_Window (PageNo, XOfs, YOfs : Integer);           External;
Function  Get_X_Offset : Integer;                              External;
Function  Get_Y_Offset : Integer;                              External;
Procedure GPrintc (CNum, X, Y, ColorF, ColorB : Integer);      External;
Procedure TGPrintc (CNum, X, Y, ColorF : Integer);             External;
Procedure Print_Str (var Text; Len, X, Y,ColorF, ColorB : Integer); External;
Procedure TPrint_Str (var Text; Len, X, Y, ColorF : Integer);  External;
Procedure Set_Display_Font (Var FontData; FontNum : Integer);  External;
Procedure Draw_Bitmap (var Image; X, Y, W, H : Integer);       External;
Procedure TDraw_Bitmap (var Image; X, Y, W, H : Integer);      External;
Procedure Copy_Page (SourcePage, DestPage : Integer);          External;
Procedure Copy_Bitmap (SourcePage, X1, Y1, X2, Y2,
          DestPage, DestX1, DestY1 : Integer);                 External;

var
  FadeToPal,
  FadePal  : Array [0..767] of Byte;
  ImageBuf : Pointer;
  YActive  : Boolean;

Function isVGA : Boolean; assembler;
asm
  mov  ax, 1A00h
  int  10h
  cmp  al, 1Ah
  je   @ok
  xor  ax, ax
  jmp  @x
@ok:
  mov  ax, 1
@x:
End;

Procedure SetMode;
var
  dummy : byte;

Begin
  case mode of
    t80x25:   asm mov ax, 03; int 10h; mov maxx, 80; mov maxy, 25 end;
    t80x28:   asm mov ax, 03; int 10h; mov ax, 1111h; xor bl, bl; int 10h
                  mov maxx, 80; mov maxy, 28 end;
    t80x50:   asm mov ax, 03; int 10h; mov ax, 1112h; xor bl, bl; int 10h
                  mov maxx, 80; mov maxy, 50 end;
    tg80x25:  Begin SetMode (t80x25, 0, 0, 0); SetCharWidth (8); Set400Lines; End;
    tg80x50:  Begin SetMode (t80x50, 0, 0, 0); SetCharWidth (8); Set400Lines; End;
    g320x200: asm mov ax, 13h; int 10h; mov maxx, 320; mov maxy, 200 end;
    g320x200x16: asm mov ax, 0Dh; int 10h; mov maxx, 320; mov maxy, 200 end;
    g640x480x16: asm mov ax, 12h; int 10h; mov maxx, 640; mov maxy, 480 end;
    g256x128: begin
                asm mov ax, 13h; int 10h end;
                portw [$3D4] := $0E11;
                portw [$3D4] := $6300;
                portw [$3D4] := $3F01;
                portw [$3D4] := $4002;
                portw [$3D4] := $8603;
                portw [$3D4] := $5004;
                portw [$3D4] := $0D07;
                portw [$3D4] := $5810;
                portw [$3D4] := $FF12;
                portw [$3D4] := $2013;
                portw [$3D4] := $0915;
              end;
    g256x256: begin
                MaxX := 256; MaxY := 256;
                asm mov ax, 13h;   int 10h
                    mov dx, 03d4h; mov al, 11h; out dx, al; inc dx
                    in  al, dx;    and al, 7fh; mov ah, al; mov al, 11h
                    dec dx;        out dx, ax
                end;
                port [$3D4] := $00; port [$3D5] := $5F;
                port [$3D4] := $01; port [$3D5] := $3F;
                port [$3D4] := $02; port [$3D5] := $40;
                port [$3D4] := $03; port [$3D5] := $82;
                port [$3D4] := $04; port [$3D5] := $4E;
                port [$3D4] := $05; port [$3D5] := $9A;
                port [$3D4] := $06; port [$3D5] := $23;
                port [$3D4] := $07; port [$3D5] := $B2;
                port [$3D4] := $08; port [$3D5] := $00;
                port [$3D4] := $09; port [$3D5] := $61;
                port [$3D4] := $10; port [$3D5] := $0A;
                port [$3D4] := $11; port [$3D5] := $AC;
                port [$3D4] := $12; port [$3D5] := $FF;
                port [$3D4] := $13; port [$3D5] := $20;
                port [$3D4] := $14; port [$3D5] := $40;
                port [$3D4] := $15; port [$3D5] := $07;
                port [$3D4] := $16; port [$3D5] := $17;
                port [$3D4] := $17; port [$3D5] := $A3;
                port [$3C4] := $01; port [$3C5] := $01;
                port [$3C4] := $04; port [$3C5] := $0E;
                port [$3CE] := $05; port [$3CF] := $40;
                port [$3CE] := $06; port [$3CF] := $05;
                dummy := port [$3DA];
                port [$3C0] := $30; port [$3C0] := $41;
                port [$3C2] := $E3;
                dummy := port [$3DA];
                port [$3C0] := $33; port [$3C0] := $00;
                asm mov dx, 03d4h; mov al, 11h; out dx, al; inc dx
                    in  al, dx;    or  al, 80h; mov ah, al; mov al,11h
                    dec dx;        out dx, ax
                end;
              end;
    g80x50:   begin
                MaxX := 80; MaxY := 50;
                asm mov ax, 13h; int 10h end;
                portw [$3C4] := $604;
                portw [$3C4] := $0F02;
                portw [$3D4] := $14;
                portw [$3D4] := $E317;
                port  [$3D4] := $09;
                port  [$3D5] := port [$3D5] and $E0 + 7;
              end;
    g80x80:   begin
                MaxX := 80; MaxY := 80;
                asm mov ax, 13h; int 10h end;
                portw [$3C4] := $604;
                portw [$3C4] := $0F02;
                portw [$3D4] := $14;
                portw [$3D4] := $E317;
                portw [$3D4] := $409;
              end;
    g80x200:  begin
                MaxX := 80; MaxY := 200;
                asm mov ax, 13h; int 10h end;
                port  [$3C4] := $04;
                port  [$3C5] := port [$3C5] and $F7;
                portw [$3C4] := $F02;
                port  [$3D4] := $14;
                port  [$3D5] := port [$3D5] and $BF;
                port  [$3D4] := $17;
                port  [$3D5] := port [$3D5] or $40;
              end;
              else if (mode >= $FF) and (mode < $200) then begin
                set_vga_modex (MODE - $FF, x_vX, x_vY, x_P);
                LogMaxX := x_vX; LogMaxY := x_vY;
                NumPages := x_P;
                MaxX := x_vX; MaxY := x_vY;
              end else Begin
                Case mode of
                  g320x240x16: Begin MaxX := 320; MaxY := 240; End;
                  g320x100:    Begin MaxX := 320; MaxY := 100; End;
                  g320x80:     Begin MaxX := 320; MaxY := 80;  End;
                  g160x100:    Begin MaxX := 160; MaxY := 100; End;
                  g160x80:     Begin MaxX := 160; MaxY := 80;  End;
                End;
              End;
  End;
  PageSize := MaxX*MaxY;
  If PageSize = 0 then PageSize := 65535;
End;

Procedure WaitSync; Assembler;
Asm
  mov  dx, 3dah
@w1: in   al, dx
     test al, 8
  jnz @w1
@w2: in   al, dx
     test al, 8
  jz @w2
End;

Procedure GetFont (var Buf);
var
  x : Integer;

Begin
  inline ($FA);
  portW [$03c4] := $0402;
  portW [$03c4] := $0704;
  portW [$03ce] := $0204;
  portW [$03ce] := $0005;
  portW [$03ce] := $0006;
  For x := 0 to 255 do
    Move (MEM [$A000:x shl 5], MEM [Seg (Buf):Ofs (Buf)+(x shl 4)], 16);
  portW [$03c4] := $0302;
  portW [$03c4] := $0304;
  portW [$03ce] := $0004;
  portW [$03ce] := $1005;
  portW [$03ce] := $0e06;
  inline ($FB);
End;

Procedure SetFont (var Buf); Assembler;
Asm
  push bp
  mov  ax, 500h
  int  10h
  mov  ax, 3
  int  10h
  mov  ax, 1110h
  mov  bh, 10h
  xor  bl, bl
  mov  cx, 256
  xor  dx, dx
  mov  es, word (Buf+2)
  mov  bp, word (Buf)
  int  10h
  pop  bp
End;

Procedure SetOneChar (var Buf; C : Byte); Assembler;
Asm
  push bp
  xor  bx, bx
  xor  dx, dx
  mov  dl, C
  mov  es, word(Buf+2)
  mov  bp, word(Buf)
  mov  ax, 1110h
  mov  bh, 16
  mov  cx, 1
  int  10h
  pop  bp
End;

Procedure SetCharSet (var Buf1, Buf2; Map1, Map2 : Integer); Assembler;
Asm
  mov  dx, 03c4h;
  mov  ax, 0402h;  out  dx, ax
  mov  ax, 0704h;  out  dx, ax
  mov  dx, 03ceh;
  mov  ax, 0204h;  out  dx, ax
  mov  ax, 0005h;  out  dx, ax
  mov  ax, 0006h;  out  dx, ax

  push ds
  mov  bx, Map1
  shl  bx, 1
  mov  bx, word ptr [CharOffsets+bx]
  lds  si, Buf1
  mov  ax, $A000
  mov  es, ax
  mov  dx, 0
@Here:
  mov  di, dx
  shl  di, 5
  add  di, bx
  mov  cx, 4
  db   66h; rep  movsw
  inc  dx
  cmp  dx, 256
  jnz  @Here
  pop  ds

  push ds
  mov  bx, Map2
  shl  bx, 1
  mov  bx, word ptr [CharOffsets+bx]
  lds  si, Buf2
  mov  dx, 0
@There:
  mov  di, dx
  shl  di, 5
  add  di, bx
  mov  cx, 4
  db   66h; rep  movsw
  inc  dx
  cmp  dx, 256
  jnz  @There
  pop  ds

  mov  dx, 03c4h;
  mov  ax, 0302h;  out  dx, ax
  mov  ax, 0304h;  out  dx, ax
  mov  dx, 03ceh;
  mov  ax, 0004h;  out  dx, ax
  mov  ax, 1005h;  out  dx, ax
  mov  ax, 0e06h;  out  dx, ax
End;


Procedure IceColor (b : Boolean); Assembler;
Asm
  mov  ax, 1003h
  mov  bl, b
  int  10h
End;

Procedure SetBorderColor (b : Byte); Assembler;
Asm
  mov  ax, 1001h
  mov  bh, b
  int  10h
End;

Procedure SetCursor (Start, Stop : Byte); Assembler;
Asm
  mov  ah, 1
  mov  ch, Start
  mov  cl, Stop
  int  10h
End;

Procedure FastWrite (x, y : Word; a : Byte; s : String); Assembler;
Asm
  dec  x
  dec  y
  mov  ax, y
  mov  cl, 5
  shl  ax, cl
  mov  di, ax
  mov  cl, 2
  shl  ax, cl
  add  di, ax
  shl  x,  1
  add  di, x
  mov  ax, 0B800h
  mov  es, ax
  xor  ch, ch
  push ds
  lds  si, s
  lodsb
  mov  cl, al
  mov  ah, a
  jcxz @end
@here:
  lodsb
  stosw
  dec  cx
  jnz  @here
@end:
  pop  ds
End;

Procedure Raster (var Buf; Col, Num : Integer); Assembler;
Asm
  push ds
  mov  ds, word (Buf+2)
  mov  si, word (Buf)
  mov  bx, 3c8h
  mov  ax, 3dah
  mov  di, ax
  mov  ax, col
  mov  cx, Num
  cli
  @mainloop:
    mov  dx, bx
    xor  al, al
    out  dx, al
    inc  dx
    outsb
    outsb
    outsb

    mov  dx, di
   @l4:
     in   al, dx
     test al, 1
    jnz  @l4
   @l5:
     in   al, dx
     test al, 1
    jz   @l5

    dec  cx
  jnz  @mainloop
  sti
  pop  ds
End;

Procedure Unchain;
Begin
  portw [$3C4] := $0604;
  portw [$3D4] := $0014;
  portw [$3D4] := $E317;
End;

Procedure Rechain;
Begin
  portw [$3C4] := $0E04;
  portw [$3C4] := $0100;
  portw [$3C4] := $0300;
  portw [$3D4] := $4014;
  portw [$3D4] := $A317;
End;

Procedure Overscan;
Begin
  port [$3C2] := $A3;
End;

Procedure ScreenOFF; assembler;
Asm
  mov  dx, 3C4h
  mov  al, 01h
  out  dx, al
  inc  dx
  in   al, dx
  or   al, 20h
  out  dx, al
End;

Procedure ScreenON; assembler;
Asm
  mov  dx, 3C4h
  mov  al, 01h
  out  dx, al
  inc  dx
  in   al, dx
  and  al, 0DFh
  out  dx, al
End;

Procedure LineRepeat (b : Byte);
Begin
  port [$3D4] := 9;
  port [$3D5] := port [$3D5] and $F0+b;
End;

Procedure DisplayWidth (b : Byte);
Begin
  port [$3D4] := $13;
  port [$3D5] := b;
End;

Procedure Set400Lines;
Begin
  port [$3C2] := (port [$3CC] and 63) or 64;
End;

Procedure WriteMode (b : byte);
Begin
  port [$3CE] := 5;
  port [$3CF] := (port [$3CF] and $FC) or (b and 3);
End;

Procedure BitPlane (b : Byte);
Begin
  portw [$3C4] := (b shl 8)+2;
End;

Procedure WriteMap (b : Byte);
Begin
  port [$3C4] := 2;
  port [$3C5] := b;
End;

Procedure SetCharMap (b1, b2 : Byte);
Begin
  portW [$3C4] := $03 + ((b1 and 3) +      ((b1 and 4) shl 2) +
                        ((b2 and 3) shl 2 + (b2 and 4) shl 3)) shl 8;
End;

Procedure PelPan (b : byte);
Begin
  port [$3C0] := 51;
  port [$3C0] := (port [$3C1] and 240) or b;
End;

Procedure SetAddress (w : Word); Assembler;
Asm
  mov  ax, w
  mov  dx, 3D4h
  mov  bh, al
  mov  al, 0Ch
  mov  bl, 0Dh
  out  dx, ax
  mov  ax, bx
  out  dx, ax
End;

Procedure SetSplit (w : Word); Assembler;
Asm
  mov dx, 3d4h
  mov ax, w
  mov bh, ah
  mov bl, ah
  and bx, 201h
  mov cl, 4
  shl bx, cl
  mov ah, al
  mov al, 18h
  out dx, ax
  mov al, 7
  cli
  out dx, al
  inc dx
  in  al, dx
  sti
  dec dx
  mov ah, al
  and ah, 0efh
  or  ah, bl
  mov al, 7
  out dx, ax
  mov al, 9
  cli
  out dx, al
  inc dx
  in  al, dx
  sti
  dec dx
  mov ah, al
  and ah, 0bfh
  shl bh, 1
  shl bh, 1
  or  ah, bh
  mov al, 9
  out dx, ax
End;

Procedure SetCharWidth (b : Byte);
var
  x, y : Integer;

Begin
  If b = 8 then y := 1 else y := $800;
  x := port [$3CC] and $F3;
  If b = 9 then x := x or 4;
  port [$3C2] := x;
  inline ($FA);
  portW [$3C4] := $0100;
  portW [$3C4] := y shl 8 + 1;
  portW [$3C4] := $0300;
  Asm
    sti
    mov  ax, 1000h
    mov  bx, y
    mov  bl, 13h
    int  10h
  End;
End;

Procedure BiosPix (x, y : integer; col : Byte); Assembler;
Asm
  mov  ah, 0Ch
  mov  bh, 1
  mov  al, col
  mov  cx, x
  mov  dx, y
  int  10h
End;

(****************************************************************************)

Procedure ClearPage (p : Pointer); assembler;
Asm
  les  di, p
  db   $66; xor cx, cx
  db   $66; xor ax, ax
  mov  cx, PageSize
  shr  cx, 2
  db   $66; rep stosw
End;

Procedure ClearPageColor (p : Pointer; c : Byte); assembler;
Asm
  les  di, p
  mov  al, c
  mov  ah, al
  push ax
  db   $66; shl ax, 16
  pop  ax
  mov  cx, PageSize
  shr  cx, 2
  db   $66; rep stosw
End;

Procedure ClearPagePart (p : Pointer; x1, y1, x2, y2 : Integer); assembler;
Asm
  les  di, p
  mov  cx, x2
  sub  cx, x1
  inc  cx
  mov  ax, y1
  mov  di, ax
  shl  di, 6
  shl  ax, 8
  add  di, ax
  add  di, x1
  mov  dx, 320
  sub  dx, cx
  cld
  mov  ax, y2
  sub  ax, y1
  inc  ax
  mov  y2, ax
  db   $66; xor ax, ax
@here:
  push cx
  shr  cx, 1
  db   0Fh, 92h, 0C3h
  shr  cx, 1
  db   $66; rep stosw
  jnc  @skip
  stosw
@skip:
  or   bl, bl
  jz   @there
  stosb
@there:
  add  si, dx
  add  di, dx
  pop  cx
  dec  y2
  jnz  @here
End;

Procedure ClearPagePartColor (p : Pointer; x1, y1, x2, y2, c : Integer); assembler;
Asm
  les  di, p
  mov  cx, x2
  sub  cx, x1
  inc  cx
  mov  ax, y1
  mov  di, ax
  shl  di, 6
  shl  ax, 8
  add  di, ax
  add  di, x1
  mov  dx, 320
  sub  dx, cx
  cld
  mov  ax, y2
  sub  ax, y1
  inc  ax
  mov  y2, ax
  mov  al, byte (c)
  mov  ah, al
  push ax
  db   $66; shl ax, 16
  pop  ax
@here:
  push cx
  shr  cx, 1
  db   0Fh, 92h, 0C3h
  shr  cx, 1
  db   $66; rep stosw
  jnc  @skip
  stosw
@skip:
  or   bl, bl
  jz   @there
  stosb
@there:
  add  si, dx
  add  di, dx
  pop  cx
  dec  y2
  jnz  @here
End;

Procedure ViewPage (p : Pointer); assembler;
Asm
  db   $66; xor cx, cx
  mov  cx, PageSize
  shr  cx, 2
  push ds
  lds  si, p
  mov  ax, 0A000h
  mov  es, ax
  xor  di, di
  db   $66; rep movsw
  pop  ds
End;

Procedure ViewPagePart (p : Pointer; x1, y1, x2, y2 : Integer); assembler;
Asm
  mov  ax, 0a000h
  mov  es, ax
  push ds
  lds  si, p
  mov  cx, x2
  sub  cx, x1
  inc  cx
  mov  ax, y1
  mov  di, ax
  shl  di, 6
  shl  ax, 8
  add  di, ax
  add  di, x1
  mov  si, di
  mov  dx, 320
  sub  dx, cx
  cld
  mov  ax, y2
  sub  ax, y1
  inc  ax
@here:
  push cx
  shr  cx, 1
  db   0Fh, 92h, 0C3h
  shr  cx, 1
  db   $66; rep movsw
  jnc  @skip
  movsw
@skip:
  or   bl, bl
  jz   @there
  movsb
@there:
  add  si, dx
  add  di, dx
  pop  cx
  dec  ax
  jnz  @here
  pop  ds
End;

(** DRAW ROUTINES ***********************************************************)

Procedure g16Pixel (x, y : Integer; c : Byte); Assembler;
Asm
  mov  dx, $3CE
  xor  al, al
  mov  ah, c
  and  ah, $0F
  out  dx, ax
  mov  ax, $0F01
  out  dx, ax
  mov  al, $08
  out  dx, al

  mov  ax, 0A000h
  mov  es, ax
  mov  bx, x
  mov  cx, bx
  mov  ax, 80
  mul  Y
  shr  bx, 3
  add  bx, ax
  and  cl, 7
  mov  al, $80
  shr  al, cl
  mov  dx, $3Cf
  out  dx, al
  mov  ah, $ff
  or   es:[bx], ah
End;

Procedure g16FillBox (x1, y1, x2, y2 : Integer; c : Byte); Assembler;
Asm
  cld
  mov  dx, 3CEh
  mov  ah, c
  mov  al, 0
  out  dx, ax
  mov  ax, 0305h
  out  dx, ax
  mov  ax, 0a000h
  mov  es, ax
  mov  ax, y1
  mov  si, y2
  sub  si, ax
  jz   @NoLines
  shl  ax, 6
  mov  di, ax
  shr  ax, 2
  add  di, ax
  mov  ax, x1
  mov  cl, al
  shr  ax, 3
  mov  bx, x2
  mov  ch, bl
  shr  bx, 3
  sub  bx, ax
  push bp
  mov  bp, bx
  add  di, ax
  and  cx, 707h
  mov  bx, 0ffffh
  shr  bl, cl
  mov  cl, ch
  xor  cl, 7
  inc  cx
  shl  bh, cl
  or   bp, bp
  jnz  @NoDub
  and  bl, bh
  jz   @NoLines2
@NoDub:
  mov  al, 8
@PutLines:
  cli
  mov  al, es:[di]
  mov  es:[di], bl
  sti
  inc  di
  mov  cx, bp
  dec  cx
  js   @NoPBL
  jz   @NoPB
  mov  al, 0ffh
  rep  stosb
@NoPB:
  cli
  mov  al, es:[di]
  mov  es:[di], bh
  sti
  inc  di
@NoPBL:
  sub  di, bp
  add  di, 79
  dec  si
  jnz  @PutLines
@NoLines2:
  pop  bp
@NoLines:
  mov  ax,0005h
  out  dx, ax
End;

Procedure g16vLine (x, y, d : Integer; c : Byte); Assembler;
Asm
  push x
  push y
  push word (c)
  call g16pixel
  mov  cx, d
  mov  di, 80
  @loop:
    add  bx, di
    or   es:[bx], ah
    dec  cx
  jnz @loop
End;

Procedure g16hLine (x, y, d : Integer; c : Byte); Assembler;
Asm
  @loop:
    push x
    push y
    push word (c)
    call g16Pixel
    inc  x
    dec  d
  jnz @loop
End;

Procedure g16Line(x1, y1, x2, y2 : Integer; c : Byte);

type
  Fixed = Record
    F : Word;
    W : Integer;
  End;

var
  dx, dy, px, py : LongInt;

Begin
  dx := 0;
  dy := 0;
  px := 0;
  py := 0;
  Fixed (px).W := x1;
  Fixed (dx).W := (x2-x1);
  Fixed (py).W := y1;
  Fixed (dy).W := (y2-y1);
  x1 := abs (Fixed (dx).W);
  If abs (Fixed (dy).W) > x1 then x1 := abs (Fixed (dy).W);
  If x1=0 then g16Pixel (x2, y2, c) else Begin
    dx := dx div x1;
    dy := dy div x1;
    Repeat
      g16Pixel (Fixed (px).W, Fixed (py).W, c);
      Inc (px, dx);
      Inc (py, dy);
      Dec (x1);
    Until x1=0;
  End;
End;

Procedure g16Ellipse (x, y, r : Integer; c : Byte);
var
  d, rr, loop : LongInt;
Begin
  rr := r*r;
  For loop := 0 to r-1 do Begin
    d := iSqrt (rr-Loop*Loop);
    g16vLine (x-Loop, y-d, d+d, c);
    g16vLine (x+loop, y-d, d+d, c);
  End;
End;

Procedure g16Clear; Assembler;
Asm
  mov  dx, $3CE
  xor  al, al
  xor  ah, ah
  out  dx, ax
  mov  ax, $0F01
  out  dx, ax
  mov  al, $08
  out  dx, al
  inc  dx
  mov  al, -1
  out  dx, al
  mov  ax, 0a000h
  mov  es, ax
  xor  di, di
  mov  ax, -1
  mov  cx, 80*800/4;
  db   $66; rep stosw;
End;

Procedure Line (x1, y1, x2, y2 : Integer; c : Byte; Page : Word);
var
  M, I, D, Dx, Dy, IncX, IncY,
                   IncA, IncB : integer;

Begin
  If Y1 < MinY then Y1 := MinY;
  If Y1 > MaxY then Y1 := MaxY;
  If Y2 < MinY then Y2 := MinY;
  If Y2 > MaxY then Y2 := MaxY;
  If X1 < MinX then X1 := MinX;
  If X1 > MaxX then X1 := MaxX;
  If X2 < MinX then X2 := MinX;
  If X2 > MaxX then X2 := MaxX;
  Dx := Abs (X2-X1);
  Dy := Abs (Y2-Y1);
  If Dx >= Dy then Begin
    If X2 > X1 then IncX := 1 else IncX := -1;
    If Y2 > Y1 then IncY := MaxX else IncY := -MaxX;
  End else Begin
    D := Dx;
    Dx := Dy;
    Dy := D;
    If X2 > X1 then IncY := 1 else IncY := -1;
    If Y2 > Y1 then IncX := MaxX else IncX := -MaxX;
  End;
  D := Dy shl 1-Dx;
  IncA := D-Dx;
  IncB := D+Dx;
  M := MaxX*Y1+X1;
  For I := 0 to Dx do Begin
    MEM [Page:M] := C;
    Inc (M, IncX);
    if D >= 0 then Begin
      Inc (D, IncA);
      Inc (M, IncY);
    End else Inc (D, IncB);
  End;
End;

Procedure Ellipse (x, y, Rad : Integer; c : Byte; Page : Word);
var
  XX4, XX, YY, D : Integer;

Begin
  XX := 0;
  YY := Rad;
  D := 3-(Rad shl 1);
  While XX <= YY do Begin
    FillChar (MEM [Page:MaxX*(Y+YY)+X-XX], XX shl 1, c);
    FillChar (MEM [Page:MaxX*(Y-YY)+X-XX], XX shl 1, c);
    FillChar (MEM [Page:MaxX*(Y+XX)+X-YY], YY shl 1, c);
    FillChar (MEM [Page:MaxX*(Y-XX)+X-YY], YY shl 1, c);
    XX4 := XX shl 2;
    If D < 0 then Inc (D, XX4+6) else Begin
      Inc (D, XX4-YY shl 2+10);
      Dec (YY);
    End;
    Inc (XX);
  End;
End;

(** PALETTE *****************************************************************)

Procedure SetRGB (c, r, g, b : Byte); Assembler;
Asm
  mov  dx, 3c8h
  mov  al, c
  out  dx, al
  inc  dx
  mov  al, r
  out  dx, al
  mov  al, g
  out  dx, al
  mov  al, b
  out  dx, al
End;

Procedure GetPalette; Assembler;
Asm
  mov  dx, 3c7h
  xor  al, al
  out  dx, al
  inc  dx; inc dx
  mov  es, Word (Buf+2)
  mov  di, Word (Buf)
  mov  cx, 768
  rep  insb
End;

Procedure SetPalette; Assembler;
Asm
  push ds
  mov  ds, Word (Buf+2)
  mov  si, Word (Buf)
  mov  dx, 3C8h
  xor  al, al
  out  dx, al
  inc  dx
  mov  cx, 768
  @baah:
    lodsb
    out dx, al
    dec cx
  jnz @baah
  pop  ds
End;

procedure BlankPalette; Assembler;
Asm
  mov  dx, 3C8h
  xor  al, al
  out  dx, al
  inc  dx
  mov  cx, 768
  @L:
    out  dx, al
    dec  cx
  jnz  @L
End;

Procedure RampPalette (var Buf; r1, g1, b1, r2, g2, b2, c1, c2 : Byte);
var
  Delta   : Integer;
  x, y, z : Integer;
  Palette : Array [0..255, 0..2] of Byte absolute Buf;

Begin
  Delta := c2 - c1;
  for x := 0 to Delta do Begin
    Palette [c1+x, 0] := r1*(Delta-x) div Delta + r2*x div Delta;
    Palette [c1+x, 1] := g1*(Delta-x) div Delta + g2*x div Delta;
    Palette [c1+x, 2] := b1*(Delta-x) div Delta + b2*x div Delta;
  End;
End;

Procedure NegatePalette (var Buf; c1, c2 : Byte);
var
  x, y, z : Integer;
  Palette : Array [0..255, 0..2] of Byte absolute Buf;

Begin
  For x := c1 to c2 do Begin
    Palette [x, 0] := 63 - Palette [x, 0];
    Palette [x, 1] := 63 - Palette [x, 1];
    Palette [x, 2] := 63 - Palette [x, 2];
  End;
End;

Procedure ScalePalette (var Buf; r, g, b, c1, c2 : Byte);
var
  x, y, z : Integer;
  Palette : Array [0..255, 0..2] of Byte absolute Buf;

Begin
  For x := c1 to c2 do Begin
    y := (Palette [x, 0] + Palette [x, 1] + Palette [x, 2]) div 3;
    Palette [x, 0] := y * r div 63;
    Palette [x, 1] := y * g div 63;
    Palette [x, 2] := y * b div 63;
  End;
End;

Procedure SlidePalette (var Buf; c1, c2 : Byte; Direction : Boolean);
var
  Save    : Longint;
  Palette : Array [0..255, 0..2] of Byte absolute Buf;

Begin
  If Direction then Begin
    Move (Palette [c2], Save, 3);
    Move (Palette [c1], Palette [c1+1], (c2-c1)*3);
    Move (Save, Palette [c1], 3);
  End else Begin
    Move (Palette [c1], Save, 3);
    Move (Palette [c1+1], Palette [c1], (c2-c1)*3);
    Move (Save, Palette [c2], 3);
  End;
End;

Procedure InitFade;
Begin
  GetPalette (FadePal);
  FadeStep := 64;
End;

Function Fadeout : Boolean; Assembler;
Asm
  mov  ax, seg FadePal
  mov  es, ax
  mov  di, offset FadePal
  mov  cx, 768
  mov  dx, 3c8h
  xor  al, al
  out  dx, al
  inc  dx
  @here:
    mov  al, es:[di]
    cmp  al, 0
    jz   @a
      dec  al
    @a:
    stosb
    out  dx, al
    dec  cx
  jnz  @here
  mov  al, FadeStep
  dec  al
  jnz  @fuck
  mov  ax, FALSE
  jmp  @exit
  @fuck:
  mov  FadeStep, al
  mov  ax, TRUE
  @exit:
End;

Procedure InitFadeTo;
Begin
  GetPalette (FadePal);
  Move (Buf, FadeToPal, 768);
  FadeStep := 64;
End;

Function FadeTo : Boolean;
var
  x, y, z : Integer;

Begin
  Port [$3C8] := 0;
  For x := 0 to 767 do Begin
    If FadePal [x] < FadeToPal [x] then Inc (FadePal [x]) else
      If FadePal [x] > FadeToPal [x] then Dec (FadePal [x]);
    Port [$3C9] := FadePal [x];
  End;
  Dec (FadeStep);
  FadeTo := (FadeStep <> 0);
End;

(** MISCELLANEOUS ***********************************************************)

Function iSqrt (x : Longint) : word; Assembler;
Asm
  db $66; xor ax, ax
  db $66; xor dx, dx
  db $66; mov di, word ptr x
  mov cx, 32
@L:
    db $66; shl di, 1
    db $66; rcl dx, 1
    db $66; shl di, 1
    db $66; rcl dx, 1
    db $66; shl ax, 1
    db $66; mov bx, ax
    db $66; shl bx, 1
    db $66; inc bx
    db $66; cmp dx, bx
    jl @S
    db $66; sub dx, bx
    db $66; inc ax
@S:
    dec cx
  jnz @L
  db $66; add ax, $8000; dw 0
  db $66; shr ax, 16
End;

Function KeyHit : Boolean; Assembler;
Asm
  mov  bx, 0040h
  mov  es, bx
  mov  ax, es:[001ch]
  sub  ax, es:[001ah]
End;

Procedure SetTimerFreq (w : Word); Assembler;
Asm
  cli
  mov  al,  00110110b
  out  43h, al
  mov  cx,  w
  mov  ax,  cx
  out  40h, al
  xchg al,  ah
  out  40h, al
  sti
End;

Function  LoadCEL (fn : String) : Pointer;
var
  f      : File;
  p      : Pointer;
  xs, ys : Integer;

Begin
  Assign (f, fn);
  {$I-} Reset (f, 1); {$I+}
  If IOResult <> 0 then Begin
    Exit;
    VGAError := -1;
  End else VGAError := 0;
  Seek (f, 4);
  Blockread (f, xs, 2);
  ys := (Filesize (f)-800) div xs;
  Seek (f, 800);
  Getmem (p, xs*ys + 4);
  LoadCEL := p;
  Word (p^) := xs; p := ptr (seg (p^), ofs (p^)+2);
  Word (p^) := ys; p := ptr (seg (p^), ofs (p^)+2);
  Blockread (f, p^, xs*ys);
  Close (f);
End;

Constructor Morph.Init;
Begin
  Size := _Size;
  Stages := _Stages;
  Fillchar (Result^, Size * Sizeof (Integer), 0);
End;

Procedure Morph.Process;
var
 i    : Integer;

Begin
  For i := 0 to Size-1 do
    Result^ [i] := (v1^ [i]- v2^ [i])*Stage div Stages+ v2^ [i];
End;

Destructor Morph.Done;
Begin
End;

Constructor BMorph.Init;
Begin
  Size := _Size;
  Stages := _Stages;
  Fillchar (Result^, Size, 0);
End;

Procedure BMorph.Process;
var
 i    : Integer;

Begin
  For i := 0 to Size-1 do
    Result^ [i] := (v1^ [i]- v2^ [i])*Stage div Stages+ v2^ [i];
End;

Destructor BMorph.Done;
Begin
End;

Constructor LMorph.Init;
Begin
  Size := _Size;
  Stages := _Stages;
  Fillchar (Result^, Size*Sizeof(LongInt), 0);
End;

Procedure LMorph.Process;
var
 i    : Integer;

Begin
  For i := 0 to Size-1 do
    Result^ [i] := (v1^ [i]- v2^ [i])*Stage div Stages+ v2^ [i];
End;

Destructor LMorph.Done;
Begin
End;

Begin
  YActive := FALSE;
  tgCount := 0;
End.