(* VGA Library ------------------------------------------------------------ *)
(*                                                                          *)
(* MODULE NAME: Autodesk FLI Module                                         *)
(* VERSION    : 0.10                                                        *)
(* LAST CHANGE: 20-01-95                                                    *)
(* AUTHOR     : BoRZoM / [TATOO]                                            *)
(* DIRECTIVES : *) {$G+,A+,E+,F+,R-,S-,D-} (*                               *)
(*                                                                          *)
(*                                                                          *)
(* NOTES      : Compatible with FLI format ONLY                             *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                          *) UNIT VGAFLI; (*

(****************************************************************************)
(******************************) INTERFACE (*********************************)
(****************************************************************************)

var
  FLIPalette : Array [0..767] of Byte;
  FLIPage    : Pointer;

Procedure InitFLI (s : String);
Procedure DeinitFLI;
Procedure ProcessFrame;
Procedure ZeroFrame;
Function  GetFrames : Word;
Function  Frame : Word;

(****************************************************************************)
(****************************) IMPLEMENTATION (******************************)
(****************************************************************************)

Const
  C_COLOR = $A+1;
  C_LC    = $A+2;
  C_BLACK = $A+3;
  C_BRUN  = $A+5;
  C_COPY  = $A+6;

Type
  FLIHeaderType = Record
     Size     : LongInt;
     Unused,
     Frames,
     Width,
     Height,
     Depth,
     Flags,
     Speed    : Word;
     Next,
     Frit     : LongInt;
     Reserved : Array [0..101] of Byte;
  End;
  FrameHeaderType = Record
    Size     : LongInt;
    Unused,
    Chunks   : Word;
    Reserved : Array [0..7] of Byte;
  End;
  ChunkHeaderType = Record
    Size      : LongInt;
    ChunkType : Word;
  End;

  TChunk = Array [0..0] of Byte;
  PChunk = ^TChunk;

var
  FLIHeader    : FLIHeaderType;
  FrameHeader  : FrameHeaderType;
  ChunkHeader  : ChunkHeaderType;
  CurrentFrame,
  ChunkOffset  : Word;
  Chunk        : PChunk;
  f            : File;

(****************************************************************************)

Function GetByte : Byte;
Begin
  GetByte := Chunk^ [ChunkOffset];
  Inc (ChunkOffset);
end;

Function GetWord : Word;
Begin
  GetWord := GetByte shl 8 + GetByte;
End;

Procedure FLIColor;
var
  Packets,
  cStart, cEnd : Byte;
  x, y, z      : Integer;

Begin
  ChunkOffset := 0;
  For x := 1 to Getword do Begin
    cStart := GetByte;
    cEnd := GetByte;
    If cEnd = 0 then cEnd := 255;
    Move (Chunk^ [ChunkOffset], FLIPalette [y*3], (cEnd-cStart+1)*3);
  End;
End;

Procedure FLILC;
var
  Packets,
  Skipcount : Byte;
  SizeCount : Shortint;
  a, b,
  x, y, z   : Integer;

Begin
  ChunkOffset := 0;
  a := 0;
  b := GetWord;
  For x := 1 to GetWord do Begin
    For y := 1 to GetByte do Begin
      Inc (a, Getbyte);
      SizeCount := GetByte;
      If SizeCount < 0 then Begin
        SizeCount := -SizeCount;
        FillChar (Mem [SEG (FLIPage^):b*320+a], SizeCount, GetByte);
        Inc (a, SizeCount);
      End else If SizeCount > 0 then Begin
        Move (Chunk^ [ChunkOffset], Mem [SEG (FLIPage^):b*320+a], SizeCount);
        Inc (ChunkOffset, SizeCount);
        Inc (a, SizeCount);
      End;
    End;
    a := 0;
    Inc (b);
  End;
End;

Procedure FLIBrun;
var
  Packets   : Byte;
  Sizecount : Shortint;
  a, b,
  x, y, z   : Integer;

Begin
  a := 0;
  b := 0;
  ChunkOffset := 0;
  For x := 1 to 200 do Begin
    Packets := GetByte;
    For y := 1 to Packets do Begin
      SizeCount := GetByte;
      If SizeCount > 0 then Begin
        FillChar (Mem [SEG (FLIPage^):b*320+a], SizeCount, GetByte);
        Inc (a, SizeCount);
      End else If SizeCount < 0 then Begin
        SizeCount := -SizeCount;
        Move (Chunk^ [ChunkOffset], Mem[SEG (FLIPage^):b*320+a], SizeCount);
        Inc (a, Sizecount);
        Inc (ChunkOffset, SizeCount);
      End;
    End;
    a := 0;
    Inc (b);
  End;
End;

Procedure FLICopy;
Begin
  Move (Chunk^, MEM [SEG (FLIPage^):0], 320*200);
End;

Procedure FLIBlack;
Begin
  FillChar (FLIPage^, 65535, 0);
End;

(****************************************************************************)

Procedure InitFLI (s : String);
Begin
  CurrentFrame := 1;
  FillChar (FLIPage^, 65535, 0);
  FillChar (FLIPalette, 768, 0);
  If Maxavail < 65535 then Exit;
  Getmem (Chunk, 65535);
  Assign (f, s);
  {$I-} Reset (f, 1); {$I+}
  If IOResult <> 0 then Exit;
  Blockread (f, FLIHeader, Sizeof (FLIHeader));
End;

Procedure DeinitFLI;
Begin
  Close (f);
  If Chunk <> NIL then Freemem (Chunk, 65535);
End;

Procedure ProcessFrame;
var
  x : Integer;

Begin
  Inc (CurrentFrame);
  If CurrentFrame > FLIHeader.Frames then ZeroFrame;
  Blockread (f, FrameHeader, Sizeof (FrameHeader));
  For x := 1 to FrameHeader.Chunks do Begin
    Blockread (f, ChunkHeader, Sizeof (ChunkHeader));
    Blockread (f, Chunk^, ChunkHeader.Size - SizeOf (ChunkHeader));
    Case ChunkHeader.ChunkType of
      C_COLOR : FLIColor;
      C_LC    : FLILC;
      C_BRUN  : FLIBrun;
      C_COPY  : FLICopy;
      C_BLACK : FLIBlack;
    End;
  End;
End;

Procedure ZeroFrame;
Begin
  CurrentFrame := 1;
  Seek (f, 0);
  Blockread (f, FLIHeader, Sizeof (FLIHeader));
End;

Function GetFrames : Word;
Begin
  GetFrames := FLIHeader.Frames;
End;

Function Frame : Word;
Begin
  Frame := CurrentFrame;
End;

End.