(* VGA Library ------------------------------------------------------------ *)
(*                                                                          *)
(* MODULE NAME: Effects Module                                              *)
(* VERSION    : 0.10                                                        *)
(* LAST CHANGE: 28-12-95                                                    *)
(* AUTHOR     : BoRZoM / [TATOO]                                            *)
(* DIRECTIVES : *) {$G+,A+,E+,F+,R-,S-,D-} (*                               *)
(*                                                                          *)
(*                                                                          *)
(* NOTES      : This unit contains high-level effects, bitmap processing    *)
(*              routines, screen fades, and all related shit.               *)
(*                                                                          *)
(*              All routines are all 320x200 liner mode compatible.         *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*                          *) UNIT VGAFX;   (*                             *)

(****************************************************************************)
(******************************) INTERFACE (*********************************)
(****************************************************************************)

(** Virtual Screen FX *******************************************************)

Procedure SmearPage (p : Pointer);
Procedure BurnPage (p : Pointer);
Procedure fx1Page (p : Pointer);
Procedure MotionClearPage (p : Pointer; Step : Byte);
Procedure Pixelate (Src, Dest : Pointer; Res : Integer);

(** Text mode fades * All-mode compatible ***********************************)

Procedure tf_Zoom (var Buf);
Procedure tf_Scroll (var Buf);
Procedure tf_Melt (var Buf);
Procedure td_Zoom;
Procedure td_Scroll;
Procedure td_Melt;

(** Graphic mode fades ******************************************************)

Procedure gf_Open  (var Buf);
Procedure gf_Zoom  (var Buf);
Procedure gf_Fall  (var Buf);
Procedure gf_Roll  (var Buf);
Procedure gf_Smear (var Buf);
Procedure gf_Pixel (var Buf);

(****************************************************************************)
(****************************) IMPLEMENTATION (******************************)
(****************************************************************************)

uses VGA;

Procedure MotionClearPage (p : Pointer; step : Byte); assembler;
asm
  les di, p
  mov bl, Step
  mov cx, PageSize
@here:
    mov al, es:[di]
    sub al, bl
    jns @skip
    xor al, al
  @skip:
    stosb
    dec cx
  jnz @here
End;

Procedure SmearPage (p : Pointer); Assembler;
Asm
  push bp
  les  di, p
  mov  bx, MaxX
@blurpixel:
  xor  ah, ah
  mov  al, es:[di+1]
  mov  dx, ax
  mov  al, es:[di-1]
  add  dx, ax
  mov  al, es:[di+bx]
  add  dx, ax
  mov  al, es:[di+320]
  add  dx, ax
  shr  dx, 2
@@xx:
  or   dl, dl
  jz   @nodec
  dec  dl
@nodec:
  mov  es:[di+320], dl
  inc  di
  mov  ax, PageSize
  sub  ax, MaxX
  cmp  di, ax
  jnz  @blurpixel
  pop  bp
End;

Procedure fx1Page (p : Pointer); Assembler;
Asm
  push bp
  les  di, p
  mov  bx, MaxX
@blurpixel:
  xor  ah, ah
  mov  al, es:[di+160]
  mov  dx, ax
  mov  al, es:[di-160]
  add  dx, ax
  mov  al, es:[di+0ffh]
  add  dx, ax
  mov  al, es:[di+0ffh]
  add  dx, ax
  shr  dx, 2
@@xx:
  or   dl, dl
  jz   @nodec
  dec  dl
@nodec:
  mov  es:[di-320], dl
  inc  di
  mov  ax, PageSize
  sub  ax, MaxX
  cmp  di, ax
  jnz  @blurpixel
  pop  bp
End;

Procedure Pixelate (Src, Dest : Pointer; Res : Integer); assembler;
var
  xAdder,
  yAdder,
  xRes   : word;

Asm
  push ds
  mov  ax, res
  mov  xadder, ax
  mov  bx, MaxX
  mul  bx
  mov  yadder, ax
  mov  cx, MaxY
  xor  dx, dx
  mov  ax, MaxX
  mov  xres, ax
  lds  si, src
  les  di, dest
  @yloop:
    push dx
    push cx
    mov  cx, XRes
    xor  bx, bx
    xor  dx, dx
    @xloop:
      mov  al, ds:[si+bx]
      stosb
      inc  dx
      mov  ax, res
      cmp  dx, ax
      jnz  @s
      add  bx, xadder
      xor  dx, dx
      @s:
      dec  cx
    jnz  @xloop
    pop  cx
    pop  dx
    inc  dx
    mov  ax, res
    cmp  dx, ax
    jnz  @s2
      add  si, yadder
      xor  dx, dx
    @s2:
    dec  cx
  jnz  @yloop
  pop  ds
End;

Procedure BurnPage (p : Pointer); Assembler;
Asm
  les  di, p
  mov  bx, MaxX
@blurpoint:
  xor  ax, ax
  xor  dx, dx
  mov  dl, es:[di]
  mov  al, es:[di+bx]
  add  dx, ax
  mov  al, es:[di-1]
  add  dx, ax
  mov  al, es:[di+1]
  add  dx, ax
  shr  dx, 2
  jz   @nodec
  dec  dx
@nodec:
  mov  es:[di-319], dl
  inc  di
  mov  ax, PageSize
  sub  ax, MaxX
  cmp  di, ax
  jb   @blurpoint
End;

(****************************************************************************)

Procedure tf_Zoom (var Buf);
var
  x, y : Integer;

Begin
  x := MEM [$40:132]*160+160;
  y := MEM [$40:133]-1;
  Fillchar (MEM [$B800:4000], 4000, 0);
  port [$3D4] := 9; port [$3D5] := 1;
  Move (Buf, MEM [$B800:$0000], x);
  For x := 1 to y do Begin
    WaitSync;
    port [$3D4] := 9; port [$3D5] := x;
    WaitSync;
  End;
End;

Procedure tf_Scroll (var Buf);
var
  x, y, z : Integer;
  l, s    : Integer;
  Buff    : Array [0..7999] of Byte absolute Buf;

Begin
  z := MEM [$40:132]+1;
  l := MEM [$40:133]-1;
  s := z*160;
  For x := 0 to z-1 do Begin
    Move (Buff [x*160], MEM [$B800:160*z], 160);
    For y := 0 to l do Begin
      port [$3D4] := 8; port [$3D5] := y;
      WaitSync;
    End;
    port [$3D4] := 8; port [$3D5] := 0;
    Move (MEM [$B800:160], MEM [$B800:$0000], s);
    port [$3D4] := 8; port [$3D5] := l;
  End;
  port [$3D4] := 8; port [$3D5] := 0;
End;

Procedure tf_Melt (var Buf); Assembler;
Asm
  push ds
  mov  ax, 40h
  mov  ds, ax
  mov  si, 132
  mov  ax, 0B800h
  mov  es, ax
  lodsb
  mov  bx, 160
  mul  bx
  mov  dx, ax
  lds  si, Buf
@start:
  xor  ax, ax
  mov  cx, dx
  xor  bx, bx
  mov  di, word (Buf)
  xor  si, si
@loop:
  mov  al, es:[di]
  cmp  al, ds:[si]
  je   @skip
  jl   @increase
  dec  al
  mov  es:[di], al
  inc  bx
  jmp  @skip
@increase:
  inc  al
  mov  es:[di], al
  inc  bx
@skip:
  mov  al, ds:[si+1]
  mov  es:[di+1], al
  add  di, 2
  add  si, 2
  dec  cx
  jnz  @loop
  cmp  bx, 0
  jnz  @start
  pop  ds
End;


Procedure td_Zoom;
var
  x, y, z : Integer;

Begin
  z := MEM [$40:132]*160+160;
  y := MEM [$40:133]-1;
  Fillchar (MEM [$B800:4000], 4000, 0);
  port [$3D4] := 9; port [$3D5] := 1;
  For x := y downto 1 do Begin
    WaitSync;
    port [$3D4] := 9; port [$3D5] := x;
    WaitSync;
  End;
  Asm
    mov ax, 0b800h
    mov es, ax
    xor di, di
    mov cx, z
    shr cx, 1
    mov al, 32
    mov ah, 07
    rep stosw
  End;
  port [$3D4] := 9; port [$3D5] := y;
End;

Procedure td_Scroll;
var
  x, y, z : Integer;
  l, s    : Integer;

Begin
  z := MEM [$40:132]+1;
  l := MEM [$40:133]-1;
  s := z*160;
  For x := 0 to z-1 do Begin
    Fillchar (MEM [$B800:160*z], 160, 0);
    For y := 0 to l do Begin
      port [$3D4] := 8; port [$3D5] := y;
      WaitSync;
    End;
    port [$3D4] := 8; port [$3D5] := 0;
    Move (MEM [$B800:160], MEM [$B800:$0000], s);
    port [$3D4] := 8; port [$3D5] := l;
  End;
  port [$3D4] := 8; port [$3D5] := 0;
End;

Procedure td_Melt; Assembler;
Asm
  push ds
  mov  ax, 40h
  mov  ds, ax
  mov  si, 132
  mov  ax, 0B800h
  mov  es, ax
  lodsb
  mov  bx, 160
  mul  bx
  mov  dx, ax
@start:
  xor  ax, ax
  mov  cx, dx
  xor  bx, bx
  mov  di, 0
@loop:
  mov  al, es:[di]
  cmp  al, 20h
  je   @skip
  jl   @increase
  dec  al
  mov  es:[di], al
  inc  bx
  jmp  @skip
@increase:
  inc  al
  mov  es:[di], al
  inc  bx
@skip:
  add  di, 2
  dec  cx
  jnz  @loop
  cmp  bx, 0
  jnz  @start
  pop  ds
End;

(****************************************************************************)

Procedure gf_Open (var Buf);
var
  x, y, z : Integer;
  y2      : Integer;

Begin
  z := 1;
  y2 := MaxY shr 1;
  While z <= Y2 do Begin
    For y := 1 to z do Begin
      Move (MEM [seg (Buf):(MaxX*y-1)], mem [$A000:MaxX*(Y2-z+y)], MaxX);
      Move (MEM [seg (Buf):(MaxX*(MaxY-y))], mem [$A000:MaxX*(Y2+z-y)], MaxX);
    End;
    WaitSync;
    z := z + 1;
  End;
End;

Procedure gf_Zoom (var Buf);
var
  x, y, z : Integer;
  y2      : Integer;
  d       : Integer;
  s, r    : Real;

Begin
  z := 1;
  Y2 := MaxY shr 1;
  While z <= Y2 do Begin
    r := 0;
    s := Y2/z;
    For y := 0 to z do Begin
      D := Round (s*r);
      Move (MEM [Seg (Buf):MaxX*(y2-d)], MEM [$A000:MaxX*(y2-y)], MaxX);
      Move (MEM [Seg (Buf):MaxX*(y2+d-1)], MEM [$A000:MaxX*(y2+y)], MaxX);
       R := R + 1;
    End;
    WaitSync;
    z := z + 1;
  End;
End;

Procedure gf_Fall (var Buf);
var
  Count,
  x, y, z : Integer;
  Speed,
  Adding,
  LineY,
  Last    : Array [1..256] of Integer;

Begin
  For z := 1 to MaxY do Begin
    Liney  [z] := MaxY+10;
    Adding [z] := 0;
    Speed  [z] := z-1;
    While Speed [z] <= MaxY do Begin
      Inc (Adding [z]);
      Inc (Speed [z], Adding [z]);
    End;
  End;
  Count := 0;
  While (Count <= MaxY) and (Liney [MaxY] < MaxY+50) do Begin
    For z := 1 to Count do Begin
      If (Liney [z]=MaxY+10) then Begin
        Liney [z] := Speed [z];
        Last [z] := 0;
      End;
      If Last [z] < MaxY then
        Fillchar (MEM [$A000:Last [z]*MaxX], MaxX, 0);
      If Liney [z] < MaxY then
        Move (MEM [Seg (Buf):(MaxX*(Z-1))],
              MEM [$A000:Liney [z]*MaxX], MaxX);
      If Liney [z] >= z then Begin
        Last [z] := Liney [z];
        Liney [z] := Liney [z] - Adding [z];
        If (Adding [z] > 0) then Dec (Adding [z]);
      End else Begin
        Liney [z] := MaxY+50;
        Last [z] := MaxY+50;
      End;
    End;
    WaitSync;
    If Count < MaxY+1 then Inc (Count);
  End;
End;

Procedure gf_Roll (var Buf);
var
  x, y, z,
  Size,
  Size2    : Integer;

Begin
  Size := 20;
  Size2 := Size*2+1;
  For x := 1 to MaxY do Begin
    Move (MEM [Seg (Buf):(x-1)*MaxX], MEM [$A000:(x-1)*MaxX], MaxX);
    For y := 0 to Size do
      If X+Size2-Y <= MaxY-1 then
        Move (MEM [Seg (Buf):(x+Size2-y)*MaxX], MEM [$A000:(x+y)*MaxX], MaxX)
          else Move (MEM [Seg (Buf):(x+y)*MaxX], MEM [$A000:(x+y)*MaxX], MaxX);
    WaitSync;
  End;
End;

Procedure gf_Smear (var Buf);
var
  x, y, z : Integer;

Begin
  For x := 0 to MaxY do Begin
    For y := MaxY downto x do
      Move (MEM [Seg (Buf):x*MaxX], MEM [$A000:y*MaxX], MaxX);
    WaitSync;
  End;
End;

Procedure gf_Pixel (var Buf);
var
  x, y, z : Integer;
Begin
  For x := 32 downto 1 do Begin
    Pixelate (@Buf, @Screen, x);
    WaitSync;
  End;
End;

(****************************************************************************)

End.